# -*- coding: utf-8 -*-

from importlib import import_module

from simmate.workflow_engine import Workflow, S3Task


ALL_WORKFLOW_TYPES = [
    "static_energy",
    "relaxation",
    "population_analysis",
    "band_structure",
    "density_of_states",
    "molecular_dynamics",
    "diffusion",
]


def get_list_of_workflows_by_type(flow_type: str):
    """
    Returns a list of all the workflows located in the given module
    """

    # Make sure the type is supported
    if flow_type not in ALL_WORKFLOW_TYPES:
        raise TypeError(
            f"{flow_type} is not allowed. Please use a workflow type from this"
            f" list: {ALL_WORKFLOW_TYPES}"
        )

    # grab the relevent module
    flow_module = import_module(f"simmate.workflows.{flow_type}")

    # This loop goes through all attributes (as strings) of the workflow module
    # and select only those that are workflow or s3tasks.
    workflow_names = []
    for attr_name in dir(flow_module):
        attr = getattr(flow_module, attr_name)
        if isinstance(attr, Workflow) or isinstance(attr, S3Task):
            # We remove the _workflow ending for each because it's repetitve.
            # We also change the name from "flow_name" to "flow-name".
            workflow_names.append(attr_name.removesuffix("_workflow").replace("_", "-"))
    # OPTIMIZE: is there a more efficient way to do this?

    return workflow_names


def get_list_of_all_workflows():
    """
    Returns a list of all the workflows of all types.

    This utility was make specifically for the CLI where we print out all
    workflow names for the user.
    """

    workflow_names = []
    for flow_type in ALL_WORKFLOW_TYPES:

        workflow_type_names = get_list_of_workflows_by_type(flow_type)

        # we add the type to the start of each workflow name and switch underscores
        # to dashes bc they are more reader-friendly
        workflow_type_names = [
            flow_type.replace("_", "-") + "/" + w.replace("_", "-")
            for w in workflow_type_names
        ]

        workflow_names += workflow_type_names

    return workflow_names



# -*- coding: utf-8 -*-

from scipy.constants import Avogadro

from simmate.toolkit import Structure as ToolkitStructure

# from pymatgen.analysis.prototypes import AflowPrototypeMatcher

from simmate.database.base_data_types import (
    DatabaseTable,
    table_column,
    Spacegroup,
)


class Structure(DatabaseTable):
    # This is an abstract class meant for inheritance
    class Meta:
        abstract = True

    base_info = ["structure_string", "source"]
    """
    The base information for this database table. All other columns can be calculated
    using the columns in this list.
    """

    structure_string = table_column.TextField()
    """
    The core structure information, which is written to a string and in a 
    compressed format using the `from_toolkit` method. To get back to our toolkit
    structure object, use the `to_toolkit` method.
    """

    # EXPERIMENTAL
    source = table_column.JSONField(blank=True, null=True)
    """
    Where the structure came from. This could be a number of things, including
    a third party id, a transformation of another structure, a creation method,
    or just a custom submission by the user.
    
    Source can be the name of another table or a python transformation.
    Source id can be thought of as the "parent structure id", which can be a
    string (mp-123), an integer (123 of same table), a list of these ([123,321]),
    or even be nothing. We make it a JSON field to account for all scenarios.
    
    EXAMPLES: (source --> source_id)
    
    - MaterialsProject --> mp-123
    - PyXtalStructure --> null
    - AtomicPurmutation --> 123
    - HereditaryMutation --> [123,124]
    - user_submission --> null
    """

    # EXPERIMENTAL
    # Where this calculation plays a role within a "nested" workflow calculation.
    # Becuase this structure can be reused by multiple workflows, we make this
    # a list of source-like objects. For example, a relaxation could be part of
    # a series of relaxations (like in StagedRelaxation) or it can be an initial
    # step of a BandStructure calculation.
    # parent_nested_calculations = table_column.JSONField(blank=True, null=True)

    """ Query-helper Info """

    nsites = table_column.IntegerField()
    """
    total number of sites in the unitcell
    """

    nelements = table_column.IntegerField()
    """
    total number of unique elements
    """

    elements = table_column.JSONField()
    """
    List of elements in the structure (ex: ["Y", "C", "F"])
    """

    chemical_system = table_column.CharField(max_length=25)
    """
    the base chemical system (ex: "Y-C-F")
    
    Note: be careful when searching for elements! Running chemical_system__includes="C"
    on this field won't do what you expect -- because it will return structures
    containing Ca, Cs, Ce, Cl, and so on. If you want to search for structures
    that contain a specific element, use elements__contains="C" instead.
    """

    density = table_column.FloatField()
    """
    Density of the crystal (g/cm^3)
    """

    density_atomic = table_column.FloatField()
    """
    Density of atoms in the crystal (atoms/Angstom^3)
    """

    volume = table_column.FloatField()
    """
    Volume of the unitcell.
    
    Note: in most cases, `volume_molar` should be used instead! This is because
    volumne is highly dependent on the symmetry and the arbitray unitcell. If 
    you are truly after small volumes of the unitcell, it is likely you really 
    just want to search by spacegroup.
    """

    volume_molar = table_column.FloatField()
    """
    Molar volume of the crystal (cm^3/mol)
    """

    # The composition of the structure formatted in various ways
    # BUG: The max length here is overkill because there are many structures
    # with 8+ elements and disordered formula (e.g. "Ca2.103 N0.98")

    formula_full = table_column.CharField(max_length=50)
    """
    The chemical formula with elements sorted by electronegativity (ex: Li4 Fe4 P4 O16)
    """

    formula_reduced = table_column.CharField(max_length=50)
    """
    The reduced chemical formula. (ex: Li4Fe4P4O16 --> LiFePO4)
    """

    formula_anonymous = table_column.CharField(max_length=50)
    """
    An anonymized formula. Unique species are arranged in ordering of 
    amounts and assigned ascending alphabets. Useful for prototyping formulas. 
    For example, all stoichiometric perovskites have anonymized_formula ABC3.
    """

    # NOTE: extra fields for the Lattice and Sites are intentionally left out
    # in order to save on overall database size. Things such as...
    #   Lattice: matrix and then... a, b, c, alpha, beta, gamma, volume
    #   Sites: abc, xyz, properties, species/element, occupancy
    # shouldn't be queried directly. If you'd like to sort structures by these
    # criteria, you can still do this in python and pandas! Just not at the
    # SQL level

    """ Relationships """
    # For the majority of Structures, you'll want to have a "source" relation that
    # indicates where the structure came from. I don't include this is the abstract
    # model but there are many ways to define it. For example it may relate to another
    # Structure table or even a Calculation. In another case, the entire Structure
    # table may have the same exact source, in which case you'd make a property!

    spacegroup = table_column.ForeignKey(Spacegroup, on_delete=table_column.PROTECT)
    """
    Spacegroup information. Points to simmate.database.base_data_types.symmetry.Spacegroup
    """

    # The AFLOW prototype that this structure maps to.
    # TODO: this will be a relationship in the future
    # prototype = table_column.CharField(max_length=50, blank=True, null=True)

    """ Model Methods """

    @classmethod
    def _from_toolkit(cls, structure, as_dict=False, **kwargs):

        # BUG: This is an old line and I can't remember why I have it. Once I
        # have implemented more unittests, consider deleting. This method is
        # ment to convert to a ToolkitStructure, but the structure should
        # already be in this format...
        structure = ToolkitStructure.from_dynamic(structure)

        # OPTIMIZE: I currently store files as poscar strings for ordered structures
        # and as CIFs for disordered structures. Both of this include excess information
        # that slightly inflates file size, so I will be making a new string format in
        # the future. This will largely be based off the POSCAR format, but will
        # account for disordered structures and all limit repeated data (such as the
        # header line, "direct", listing each element/composition, etc.).
        storage_format = "POSCAR" if structure.is_ordered else "CIF"

        # OPTIMIZE
        # This attempts to match the structure to an AFLOW prototype and it is
        # by far the slowest step of loading structures to the database. Try
        # to optimize this in the future.
        # prototype = AflowPrototypeMatcher().get_prototypes(structure)
        # prototype_name = prototype[0]["tags"]["mineral"] if prototype else None

        # Given a pymatgen structure object, this will return a database structure
        # object, but will NOT save it to the database yet. The kwargs input
        # is only if you inherit from this class and add extra fields.
        structure_dict = dict(
            structure_string=structure.to(fmt=storage_format),
            nsites=structure.num_sites,
            nelements=len(structure.composition),
            elements=[str(e) for e in structure.composition.elements],
            chemical_system=structure.composition.chemical_system,
            density=structure.density,
            density_atomic=structure.num_sites / structure.volume,
            volume=structure.volume,
            # 1e-27 is to convert from cubic angstroms to Liter and then 1e3 to
            # mL. Therefore this value is in mL/mol
            # OPTIMIZE: move this to a class method
            volume_molar=(structure.volume / structure.num_sites)
            * Avogadro
            * 1e-27
            * 1e3,
            spacegroup_id=structure.get_space_group_info(0.1)[1],  # OPTIMIZE
            formula_full=structure.composition.formula,
            formula_reduced=structure.composition.reduced_formula,
            formula_anonymous=structure.composition.anonymized_formula,
            # prototype=prototype_name,
            **kwargs,  # this allows subclasses to add fields with ease
        )
        # If as_dict is false, we build this into an Object. Otherwise, just
        # return the dictionary
        return structure_dict if as_dict else cls(**structure_dict)

    def to_toolkit(self):
        """
        Converts the database object to pymatgen structure object
        """

        # NOTE: if you know this is what you're going to do from a query, then
        # it is more efficient to only grab the structure_string column because
        # that's all you need! You'd do that like this:
        #   structure_db = Structure.objects.only("structure_string").get(id="example-id")
        # This grabs the proper Structure entry and only the structure column.

        # convert the stored string to python dictionary.
        # OPTIMIZE: see my comment on storing strings in the from_toolkit method above.
        # For now, I need to figure out if I used "CIF" or "POSCAR" and read the structure
        # accordingly. In the future, I can just assume my new format.
        # If the string starts with "#", then I know that I stored it as a "CIF".
        storage_format = "CIF" if (self.structure_string[0] == "#") else "POSCAR"

        # convert the string to pymatgen Structure object
        structure = ToolkitStructure.from_str(
            self.structure_string,
            fmt=storage_format,
        )

        return structure


# TODO:
# Explore polymorphic relations instead of a JSON dictionary.
# Making relationships to different tables makes things difficult to use, so
# these columns are just standalone.
#
# This is will be very important for "source" and "parent_nested_calculations"
# fields because I have no way to efficiently convert these fields to the objects
# that they refer to. There's also no good way to access a structure's "children"
# (i.e. structure where they are the source).
#
# I should investigate generic relations in django though:
# https://docs.djangoproject.com/en/3.2/ref/contrib/contenttypes/#generic-relations
#
# Another option is using django-polymorphic.
# https://django-polymorphic.readthedocs.io/en/latest/
# This thread is really helpful on the subject:
# https://stackoverflow.com/questions/30343212/

# TODO:
# Consider adding some methods to track the history of a structure. This
# would be useful for things like evolutionary algorithms.
# get_source_parent:
#   this would iterate through sources until we find one in the same table
#   as this one. Parent sources are often the most recent transformation
#   or mutation applied to a structure, such as a MirrorMutation.
# get_source_seed:
#   this would iterate through sources until we hit a dead-end. So the seed
#   source would be something like a third-party database, a method that
#   randomly create structures, or a prototype.
# Both of these get more complex when we consider transformation that have
# multiple parents (and therefore multiple seeds too). An example of this
# is the HereditaryMutation.



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import (
    table_column,
    Structure,
    Forces,
    Thermodynamics,
    Calculation,
)


class StaticEnergy(Structure, Thermodynamics, Forces, Calculation):

    base_info = (
        [
            "valence_band_maximum",
            "conduction_band_minimum",
            "energy_fermi",
            "is_gap_direct",
        ]
        + Structure.base_info
        + Thermodynamics.base_info
        + Forces.base_info
        + Calculation.base_info
    )
    """
    The base information for this database table. All other columns can be calculated
    using the columns in this list.
    """

    # Note: we assume that only converged data is being stored! So there is no
    # "converged_electronic" section here. ErrorHandlers and workups should
    # ensure this.

    # This data here is something we only get for the final structure, so it
    # may make sense to move this data into the IonicStepStructure table (and
    # allow null values for non-final steps). I instead keep this here because
    # I don't want columns above that are largely empty.
    # Note: all entries are optional because there is no guaruntee the calculation
    # finishes successfully
    band_gap = table_column.FloatField(blank=True, null=True)
    is_gap_direct = table_column.BooleanField(blank=True, null=True)
    energy_fermi = table_column.FloatField(blank=True, null=True)
    conduction_band_minimum = table_column.FloatField(blank=True, null=True)
    valence_band_maximum = table_column.FloatField(blank=True, null=True)

    def update_from_vasp_run(self, vasprun, corrections, directory):
        # Takes a pymatgen VaspRun object, which is what's typically returned
        # from a simmate VaspTask.run() call.

        # The data is actually easier to access as a dictionary and everything
        # we need is stored under the "output" key.
        data = vasprun.as_dict()["output"]
        # In a static energy calculation, there is only one ionic step so we
        # grab that up front.
        ionic_step = data["ionic_steps"][0]

        # Take our structure, energy, and forces to build all of our other
        # fields for this datatable
        # OPTIMIZE: this overwrites structure data, which should already be there.
        # Is there a faster way to grab this data and update attributes?
        new_kwargs = self.from_toolkit(
            structure=vasprun.final_structure,
            energy=ionic_step["e_wo_entrp"],
            site_forces=ionic_step["forces"],
            lattice_stress=ionic_step["stress"],
            as_dict=True,
        )
        for key, value in new_kwargs.items():
            setattr(self, key, value)

        # There is also extra data for the final structure that we save directly
        # in the relaxation table. We use .get() in case the key isn't provided
        self.band_gap = data.get("bandgap")
        self.is_gap_direct = data.get("is_gap_direct")
        self.energy_fermi = data.get("efermi")
        self.conduction_band_minimum = data.get("cbm")
        self.valence_band_maximum = data.get("vbm")

        # lastly, we also want to save the corrections made and directory it ran in
        self.corrections = corrections
        self.directory = directory

        # Now we have the relaxation data all loaded and can save it to the database
        self.save()

    """ Set as Abstract Model """
    # I have other models inherit from this one, while this model doesn't need
    # its own table.
    class Meta:
        abstract = True
        app_label = "local_calculations"



# -*- coding: utf-8 -*-

from rest_framework.serializers import HyperlinkedModelSerializer
from rest_framework.viewsets import ReadOnlyModelViewSet

from simmate.database.base_data_types import Spacegroup
from simmate.database.third_parties import (
    AflowStructure,
    CodStructure,
    JarvisStructure,
    MatProjStructure,
    OqmdStructure,
)


"""

Instead of making an individual view for each model and each of its properties,
we instead let djangorestframework do the heavy lifting for us -- it creates
all views for us via a "Serializer" and "ViewSet". So this is a two-step process
for each model. Here is an example:

from some.package.models import MyExampleModel    

class MyExampleSerializer(HyperlinkedModelSerializer):
    class Meta:
        model = MyExampleStructure
        fields = "__all__"    

class MyExampleViewSet(ReadOnlyModelViewSet):
    queryset = MyExampleModel.objects.all()
    serializer_class = MyExampleSerializer
    filterset_fields = "__all__"

"""

# --------------------------------------------------------------------------------------


class SpacegroupSerializer(HyperlinkedModelSerializer):
    class Meta:
        model = Spacegroup
        fields = "__all__"


class SpacegroupViewSet(ReadOnlyModelViewSet):
    queryset = Spacegroup.objects.order_by("number").all()
    serializer_class = SpacegroupSerializer
    filterset_fields = "__all__"


# --------------------------------------------------------------------------------------


class MatProjSerializer(HyperlinkedModelSerializer):
    class Meta:
        model = MatProjStructure
        fields = "__all__"


class MatProjViewSet(ReadOnlyModelViewSet):
    queryset = MatProjStructure.objects.all()
    serializer_class = MatProjSerializer
    filterset_fields = "__all__"


# --------------------------------------------------------------------------------------


class JarvisSerializer(HyperlinkedModelSerializer):
    class Meta:
        model = JarvisStructure
        fields = "__all__"


class JarvisViewSet(ReadOnlyModelViewSet):
    queryset = JarvisStructure.objects.all()
    serializer_class = JarvisSerializer
    filterset_fields = "__all__"


# --------------------------------------------------------------------------------------


class AflowSerializer(HyperlinkedModelSerializer):
    class Meta:
        model = AflowStructure
        fields = "__all__"


class AflowViewSet(ReadOnlyModelViewSet):
    queryset = AflowStructure.objects.all()
    serializer_class = AflowSerializer
    filterset_fields = "__all__"


# --------------------------------------------------------------------------------------


class OqmdSerializer(HyperlinkedModelSerializer):
    class Meta:
        model = OqmdStructure
        fields = "__all__"


class OqmdViewSet(ReadOnlyModelViewSet):
    queryset = OqmdStructure.objects.all()
    serializer_class = OqmdSerializer
    filterset_fields = "__all__"


# --------------------------------------------------------------------------------------


class CodSerializer(HyperlinkedModelSerializer):
    class Meta:
        model = CodStructure
        fields = "__all__"


class CodViewSet(ReadOnlyModelViewSet):
    queryset = CodStructure.objects.all()
    serializer_class = CodSerializer
    filterset_fields = "__all__"


# --------------------------------------------------------------------------------------



# -*- coding: utf-8 -*-

# TODO: when I add more calculators, I can do something like this...
# if "simmate.calculators.vasp" in installed_apps:

# These are tasks that are in early development and don't have databases tables.
# Therefore, they are s3tasks only - not workflows.
from simmate.calculators.vasp.tasks.pre_bader import MatProjPreBaderTask

matproj_workflow = MatProjPreBaderTask()



# -*- coding: utf-8 -*-

from simmate.utilities import get_doc_from_readme

__doc__ = get_doc_from_readme(__file__)

from .tasks.supervised_staged_shell_task import S3Task
from .tasks.workflow_task import WorkflowTask
from .workflow import Workflow, Parameter, ModuleStorage
from .error_handler import ErrorHandler

# from .utilities import s3task_to_workflow  # NOT MODULAR - REQUIRES DB CONNECTION



# -*- coding: utf-8 -*-

from django.shortcuts import render, redirect
from django.contrib.auth.decorators import login_required

from prefect.backend.flow_run import FlowView

from simmate.database.local_calculations import (
    MITRelaxation,
    # MITIonicStep,
)
from simmate.website.local_calculations.forms import MITRelaxationForm
from simmate.workflow_engine import WorkflowTask


@login_required
def all_local_calculations(request):

    # now let's put the data and template together to send the user
    context = {"active_tab_id": "workflows"}
    template = "local_calculations/all.html"
    return render(request, template, context)


@login_required
def relaxations(request):

    # In order to make links to the monitoring pages for each of these, we need
    # to grab the prefect id
    # BUG: If no flow exists, a ValueError is raised. I may want to account for
    # this when the user doesn't have Prefect set up.
    mit_flow_id = FlowView.from_flow_name("MIT Relaxation").flow_id

    # now let's put the data and template together to send the user
    context = {
        "active_tab_id": "workflows",
        "mit_flow_id": mit_flow_id,
    }
    template = "local_calculations/relaxations.html"
    return render(request, template, context)


@login_required
def mit_about(request):

    # In order to make links to the monitoring pages for each of these, we need
    # to grab the prefect id
    # BUG: If no flow exists, a ValueError is raised. I may want to account for
    # this when the user doesn't have Prefect set up.
    mit_flow_id = FlowView.from_flow_name("MIT Relaxation").flow_id

    # TODO: grab some metadata about this calc. For example...
    # ncalculations = MITRelaxation.objects.count()

    # now let's put the data and template together to send the user
    context = {
        "active_tab_id": "workflows",
        "mit_flow_id": mit_flow_id,
    }
    template = "local_calculations/mit/about.html"
    return render(request, template, context)


@login_required
def mit_submit(request):

    if request.method == "POST":
        submission_form = MITRelaxationForm(request.POST, request.FILES)
        if submission_form.is_valid():
            # grab the structure (as a pymatgen object) and all other inputs
            structure = submission_form.cleaned_data["structure_file"]
            labels = submission_form.cleaned_data["labels"]
            vasp_command = submission_form.cleaned_data["vasp_command"]

            # We can now submit the workflow for the structure.
            flowtask = WorkflowTask(workflow_name="MIT Relaxation")
            flow_run_id = flowtask.run(
                structure=structure,
                vasp_command=vasp_command,
                labels=labels,
                wait_for_run=False,
            )

            return redirect(f"https://cloud.prefect.io/simmate/flow-run/{flow_run_id}")
    else:
        submission_form = MITRelaxationForm()

    # now let's put the data and template together to send the user
    context = {
        "active_tab_id": "workflows",
        "submission_form": submission_form,
    }
    template = "local_calculations/mit/submit.html"
    return render(request, template, context)


@login_required
def mit_all(request):

    # Grab the most recent 50 MIT relaxations that have been registered/ran
    calculations = MITRelaxation.objects.order_by("created_at").reverse().all()[:50]

    # We also want to count how many total calculations there are.
    ncalculations_possible = MITRelaxation.objects.count()

    # now let's put the data and template together to send the user
    context = {
        "active_tab_id": "workflows",
        "calculations": calculations,
        "ncalculations_possible": ncalculations_possible,
    }
    template = "local_calculations/mit/all.html"
    return render(request, template, context)


@login_required
def mit_single(request, mitrelax_id):

    # Grab the most recent 50 MIT relaxations that have been registered/ran
    calculation = MITRelaxation.objects.get(id=mitrelax_id)

    # Make the convergence figure and convert it to an html div
    figure_convergence = calculation.get_convergence_plot()
    figure_convergence_html = figure_convergence.to_html(
        full_html=False, include_plotlyjs=False
    )

    # now let's put the data and template together to send the user
    context = {
        "active_tab_id": "workflows",
        "calculation": calculation,
        "figure_convergence_html": figure_convergence_html,
    }
    template = "local_calculations/mit/single.html"
    return render(request, template, context)



# -*- coding: utf-8 -*-

import os

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class IncorrectShift(ErrorHandler):
    """
    This handler addresses issues in the K-point mesh that can be fixed by switching
    to a gamma-centered mesh.
    """

    # run this while the VASP calculation is still going
    is_monitor = True

    # we assume that we are checking the vasp.out file
    filename_to_check = "vasp.out"

    # These are the error messages that we are looking for in the file
    possible_error_messages = [
        "Routine TETIRR needs special values",
        "Could not get correct shifts",
    ]

    def correct(self, directory):

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # Change to Gamma-centered mesh
        incar["KGAMMA"] = True

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return "switched KGAMMA to True"



# -*- coding: utf-8 -*-

from django import forms

from simmate.toolkit import Structure


class MITRelaxationForm(forms.Form):

    # Max length refers to the filename length, not its contents
    structure_file = forms.FileField(max_length=100)

    # The vasp_command to submit with
    vasp_command = forms.CharField(initial="mpirun -n 4 vasp_std > vasp.out")

    # List of labels to submit the workflow with. Use commas to separate.
    labels = forms.CharField(initial="WarWulf")

    def clean_structure_file(self):
        # Grab what the user submitted
        structure_file = self.cleaned_data["structure_file"]

        # Make sure the file isn't too large
        if structure_file:
            # 5e6 is 5MB limit
            if structure_file.size > 5e6:
                raise Exception("File upload is too large (>5MB)")
        else:
            raise Exception("Couldn't read uploaded image")

        # If we make it through the checks above, we can now convert the
        # text of the file into a pymatgen structure object
        structure_str = structure_file.read().decode("utf-8")
        structure = Structure.from_str(structure_str, fmt="POSCAR")

        return structure

    def clean_labels(self):

        # Grab what the user submitted
        labels_str = self.cleaned_data["labels"]

        # convert the string to a list of labels.
        # For example...
        #   "WarWulf, LongLeaf, DogWood"
        #       converts to
        #   ['WarWulf', ' LongLeaf', ' DogWood']
        labels = labels_str.strip().split(",")

        return labels



# -*- coding: utf-8 -*-

import os
import json
import shutil

from pymatgen.io.vasp.outputs import Outcar

from simmate.toolkit import Structure
from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar
from simmate.calculators.vasp.outputs import Oszicar

# !!! I really don't like how all of these errors are amassed into one class because
# it makes the logic difficult to follow and it raises questions about how to handle
# multiple errors showing up at once. I still convert this class from custodian,
# but in an admittedly lazy way (there aren't enough explanations for each correction).
# I'm going to revisit this and breakdown this class into smaller ones.


class MainVaspErrors(ErrorHandler):
    """
    A centralized class that handles a number of common errors that occur
    during VASP runs.
    """

    # run this while the VASP calculation is still going
    is_monitor = True

    # These are the error messages that we are looking for in the vasp.out file.
    # Note, that we can pick which ones we search for and which we ignore. See
    # the init for more on this.
    # Also note that some errors have multiple possible messages accossiated with
    # them (such as "tet").
    all_error_messages = {
        "tet": [
            "Tetrahedron method fails",
            "Fatal error detecting k-mesh",
            "Fatal error: unable to match k-point",
            "Routine TETIRR needs special values",
            "Tetrahedron method fails (number of k-points < 4)",
            "BZINTS",
        ],
        "inv_rot_mat": ["rotation matrix was not found (increase SYMPREC)"],
        "brmix": ["BRMIX: very serious problems"],
        "subspacematrix": ["WARNING: Sub-Space-Matrix is not hermitian in DAV"],
        "tetirr": ["Routine TETIRR needs special values"],
        "incorrect_shift": ["Could not get correct shifts"],
        "real_optlay": [
            "REAL_OPTLAY: internal error",
            "REAL_OPT: internal ERROR",
        ],
        "rspher": ["ERROR RSPHER"],
        "dentet": ["DENTET"],
        "too_few_bands": ["TOO FEW BANDS"],
        "triple_product": ["ERROR: the triple product of the basis vectors"],
        "rot_matrix": [
            "Found some non-integer element in rotation matrix",
            "SGRCON",
        ],
        "brions": ["BRIONS problems: POTIM should be increased"],
        "pricel": ["internal error in subroutine PRICEL"],
        "zpotrf": ["LAPACK: Routine ZPOTRF failed"],
        "amin": ["One of the lattice vectors is very long (>50 A), but AMIN"],
        "zbrent": ["ZBRENT: fatal internal in", "ZBRENT: fatal error in bracketing"],
        "pssyevx": ["ERROR in subspace rotation PSSYEVX"],
        "eddrmm": ["WARNING in EDDRMM: call to ZHEGV failed"],
        "edddav": ["Error EDDDAV: Call to ZHEGV failed"],
        "grad_not_orth": ["EDWAV: internal error, the gradient is not orthogonal"],
        "nicht_konv": ["ERROR: SBESSELITER : nicht konvergent"],
        "zheev": ["ERROR EDDIAG: Call to routine ZHEEV failed!"],
        "elf_kpar": ["ELF: KPAR>1 not implemented"],
        "elf_ncl": ["WARNING: ELF not implemented for non collinear case"],
        "rhosyg": ["RHOSYG"],
        "posmap": ["POSMAP"],
        "point_group": ["group operation missing"],
        "symprec_noise": [
            "determination of the symmetry of your systems shows a strong"
        ],
    }

    def __init__(
        self,
        natoms_large_cell=100,
        errors_to_catch=None,
        errors_to_ignore=None,
    ):

        # The threshold number of atoms to treat the cell as a large supercell.
        # This affects the handling of some errors.
        self.natoms_large_cell = natoms_large_cell

        # In some cases, we may not want to look for all the errors that this class
        # supports. In that case, we can give a list of the ones we want to look
        # for as well as a list of the ones we want to ignore.
        # If errors_to_catch is left as None, we look for all of them
        self.errors_to_catch = errors_to_catch or list(self.error_messages.keys())
        # Then remove any that were listed in errors_to_ignore
        for error in errors_to_ignore:
            self.errors_to_catch.pop(error)

    def check(self, dir):
        """
        Check for errors in the specified directory. Note, we assume that we are
        checking the vasp.out file. If that file is not present, we say that there
        is no error because another handler will address this.
        """

        # We want to return a list of errors that were found, so we keep a
        # master list.
        errors_found = []

        # establish the full path to the output file
        filename = os.path.join(dir, "vasp.out")

        # check to see that the file is there first
        if os.path.exists(filename):

            # read the file content and then close it
            with open(filename) as file:
                file_text = file.read()

            # Check if each error is present
            for error, message in self.error_messages.items():
                # if the error is NOT present, find() returns a -1
                if file_text.find(message) != -1:

                    # SPECIAL CASE: For brmix, we sometimes want to ignore this
                    if error == "brmix":
                        # load the INCAR file to view the current settings
                        incar_filename = os.path.join(dir, "INCAR")
                        incar = Incar.from_file(incar_filename)

                        # if NELECT is in the INCAR, that means we are running
                        # a charged calculation (e.g. defects). If this is the
                        # case, then we want to ingore a change in electron
                        # density (brmix) and move on to checking the next error.
                        if "NELECT" in incar:
                            continue

                    # add to our list of errors found
                    errors_found.append(error)

        # If the file doesn't exist, we are not seeing any error yetm which is
        # also an empty list. Otherwise return the list of errors we found
        return errors_found

    def correct(self, error, dir):
        """
        Perform corrections based on the INCAR and the list of errors returned by
        the check() method.

        Some errors have a series of fixes, where we want to keep a tally of
        how many times we tried fixing it. To keep track of these, a file named
        simmate_error_counts.json is written and updated.
        """

        # Note "error" here is a list of the errors found. For example, error
        # could be ["tet", "brmix", "rhosyg"]. If there weren't any errors found
        # then this is just and empty list.

        # Multiple corrections could be applied here so we record all of them.
        corrections = []

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(dir, "INCAR")
        incar = Incar.from_file(incar_filename)

        # load the error-count file if it exists
        error_count_filename = os.path.join(dir, "simmate_error_counts.json")
        if os.path.exists(error_count_filename):
            with open(error_count_filename) as error_count_file:
                error_counts = json.load(error_count_file)
        # otherwise we are starting with an empty dictionary
        else:
            error_counts = {}

        # !!! why are these separate errors if they are always handled together?
        if "tet" in error and "dentet" in error:

            # check what the current KSPACING is. If it's not set, that means it's using
            # the default which is 0.5.
            current_kspacing = incar.get("KSPACING", 0.5)

            # If KSPACING isn't the default value, we try decreasing KSPACING
            # by 20% in each direction. which approximately doubles the number
            # of kpoints.
            if current_kspacing != 0.5:
                new_kspacing = current_kspacing * 0.8
                incar["KSPACING"] = new_kspacing
                corrections.appned(
                    f"switched KSPACING from {current_kspacing} to {new_kspacing}"
                )

            # otherwise we try changing the smearing method to guassian
            else:
                incar["ISMEAR"] = 0
                incar["SIGMA"] = 0.05
                corrections.append("switched ISMEAR to 0 and SIGMA to 0.05")

        if "inv_rot_mat" in error:
            incar["SYMPREC"] = 1e-8
            corrections.append("switched SYMPREC to 1e-8")

        if "brmix" in error:

            # The fix is based on the number of times we've already tried to
            # fix brmix. So let's make sure it's in our error_count dictionary.
            # If it isn't there yet, set the count to 0 and we'll update it below.
            error_counts["brmix"] = error_counts.get("brmix", 0)

            # The KGAMMA mode is relevent in a number of fixes below, so we
            # just grab that here. The default value is True.
            current_kgamma = incar.get("KGAMMA", True)

            # check if there is a valid OUTCAR
            if error_counts["brmix"] == 0:
                outcar_filename = os.path.join(dir, "OUTCAR")
                try:
                    assert Outcar(outcar_filename).is_stopped is False
                except Exception:
                    # if the OUTCAR isn't valid, we want to skip the first attempted
                    # correction below. We do this by adding 1 to our error count.
                    error_counts["brmix"] += 1

            if error_counts["brmix"] == 0:
                # If the OUTCAR is valid - simply rerun the job and increment
                # error count for next time.
                incar["ISTART"] = 1
                corrections.append("switched ISTART to 1")
                error_counts["brmix"] += 1

            elif error_counts["brmix"] == 1:
                # Use Kerker mixing w/default values for other parameters
                incar["IMIX"] = 1
                corrections.append("switched IMIX to 1")
                error_counts["brmix"] += 1

            elif error_counts["brmix"] == 2 and current_kgamma == True:
                # switch to Monkhorst and turn off Kerker mixing (IMIX=1)
                incar["KGAMMA"] = False
                corrections.append("switched KGAMMA to False")
                incar.pop("IMIX")
                corrections.append("removed any IMIX tag")
                error_counts["brmix"] += 1

            elif error_counts["brmix"] in [2, 3] and current_kgamma == False:
                # switch to Gamma and turn off Kerker mixing (IMIX=1)
                incar["KGAMMA"] = True
                corrections.append("switched KGAMMA to True")
                incar.pop("IMIX")
                corrections.append("removed any IMIX tag")
                error_counts["brmix"] += 1

                # TODO: I'm not sure what Custodian did here, so I haven't adapted
                # it yet. How can the number of kpoints be less than 1...?
                #
                # if vi["KPOINTS"].num_kpts < 1:
                #     all_kpts_even = all(bool(n % 2 == 0) for n in vi["KPOINTS"].kpts[0])
                #     if all_kpts_even:
                #         new_kpts = (tuple(n + 1 for n in vi["KPOINTS"].kpts[0]),)
                #         actions.append(
                #             {
                #                 "dict": "KPOINTS",
                #                 "action": {"_set": {"kpoints": new_kpts}},
                #             }
                #         )

            else:

                # Try turning off symmetry and using a Gamma-packed grid
                incar["ISYM"] = 0
                corrections.append("switched ISYM to 0")
                incar["KGAMMA"] = True
                corrections.append("switched KGAMMA to True")

                # Check the current ICHARG setting, where default is 0
                # If the ICHARG is less than 10, then we want to delete the CHGCAR
                # and WAVECAR to ensure the next run is a clean start.
                current_icharg = incar.get("ICHARG", 0)
                if current_icharg < 10:
                    os.remove(os.path.join(dir, "CHGCAR"))
                    os.remove(os.path.join(dir, "WAVECAR"))
                corrections.append("deleted CHGCAR and WAVECAR")

                # BUG: why doesn't custodian add an attempt here?
                # error_counts["brmix"] += 1

        if "zpotrf" in error:

            # We can learn more about the cause of this error by looking at
            # the OSZICAR and checking the number of steps completed.
            # BUG: what if the OSCZICAR is improperly formatted but a number
            # of ionic steps completed successfully?
            try:
                oszicar_filename = os.path.join(dir, "OSZICAR")
                oszicar = Oszicar(oszicar_filename)
                nionic_steps = len(oszicar.ionic_steps)
            except Exception:
                nionic_steps = 0

            # If the error shows up after the first ionic step, then this is likely
            # caused by POTIM being too large.
            if nionic_steps >= 1:
                # Halve the POTIM and turn off symmetry
                current_potim = incar.get("POTIM", 0.5)
                new_potim = current_potim / 2
                incar["POTIM"] = new_potim
                corrections.append(f"switch POTIM from {current_potim} to {new_potim}")
                incar["ISYM"] = 0
                corrections.append("set ISYM to 0")

            # If this is a static calculation (NSW=0) or if ISIF is 1-3, then
            # we just want to turn off symmetry
            elif incar.get("NSW", 0) == 0 or incar.get("ISIF", 0) in [0, 1, 3]:
                incar["ISYM"] = 0
                corrections.append("set ISYM to 0")

            # If the error shows up on the first ionic step, it is likely that
            # our bond distances are too small. A simple fix is to scale our
            # starting lattice. We also save the original structure to a file
            # name POSCAR_original in case it's needed elsewhere.
            else:
                poscar_filename = os.path.join(dir, "POSCAR")
                structure = Structure.from_file(poscar_filename)
                structure.to("POSCAR", poscar_filename + "_original")
                structure.apply_strain(0.2)
                structure.to("POSCAR", poscar_filename)
                corrections.append("scaled the structure lattice by +20%")

            # Check the current ICHARG setting, where default is 0
            # If the ICHARG is less than 10, then we want to delete the CHGCAR
            # and WAVECAR to ensure the next run is a clean start.
            current_icharg = incar.get("ICHARG", 0)
            if current_icharg < 10:
                os.remove(os.path.join(dir, "CHGCAR"))
                os.remove(os.path.join(dir, "WAVECAR"))
            corrections.append("deleted CHGCAR and WAVECAR")

        if "subspacematrix" in error:

            # The fix is based on the number of times we've already tried to
            # fix this. So let's make sure it's in our error_count dictionary.
            # If it isn't there yet, set the count to 0.
            # Also increase the count by 1, as this is our new addition
            error_counts["subspacematrix"] = error_counts.get("subspacematrix", 0) + 1

            # On our first try, change evalutating projection operators
            # in reciprocal space.
            if error_counts["subspacematrix"] == 0:
                incar["LREAL"] = False
                corrections.append("set LREAL to False")
            # As a last resort, try changing the precision
            else:
                incar["PREC"] = "Accurate"
                corrections.append("set PREC to Accurate")

        if error.intersection(["rspher", "real_optlay", "nicht_konv"]):

            # The fix is based on the number of times we've already tried to
            # fix brmix. So let's make sure it's in our error_count dictionary.
            # If it isn't there yet, set the count to 0 and we'll update it below.
            error_counts["real_optlay"] = error_counts.get("real_optlay", 0)

            poscar_filename = os.path.join(dir, "POSCAR")
            structure = Structure.from_file(poscar_filename)

            if structure.num_sites < self.natoms_large_cell:
                incar["LREAL"] = False
                corrections.append("set LREAL to False")

            else:
                # for large supercell, try an in-between option LREAL = True
                # prior to LREAL = False
                if error_counts["real_optlay"] == 0:
                    # use real space projectors generated by pot
                    incar["LREAL"] = True
                    corrections.append("set LREAL to True")
                elif error_counts["real_optlay"] == 1:
                    incar["LREAL"] = False
                    corrections.append("set LREAL to False")
                error_counts["real_optlay"] += 1

        if error.intersection(["tetirr", "incorrect_shift"]):
            incar["KGAMMA"] = True
            corrections.append("switched KGAMMA to True")

        if "rot_matrix" in error:
            incar["KGAMMA"] = True
            corrections.append("switched KGAMMA to True")
            # TODO: 2nd attempt is to turn ISYM=0

        if "amin" in self.errors:
            incar["AMIN"] = 0.01
            corrections.append("switched AMIN to 0.01")

        if "triple_product" in self.errors:
            poscar_filename = os.path.join(dir, "POSCAR")
            structure = Structure.from_file(poscar_filename)
            structure.to("POSCAR", poscar_filename + "_original")
            structure.make_supercell([[1, 0, 0], [0, 0, 1], [0, 1, 0]])
            structure.to("POSCAR", poscar_filename)
            corrections.append("adjusted lattice basis to swap b and c vectors")

        if "pricel" in error:
            incar["SYMPREC"] = 1e-8
            corrections.append("switched SYMPREC to 1e-8")
            incar["ISYM"] = 0
            corrections.append("switched ISYM to 0")

        if "brions" in error:
            current_potim = incar.get("POTIM", 0.5)
            new_potim = current_potim + 0.1
            incar["POTIM"] = new_potim
            corrections.append("switched POTIM from {current_potim} to {new_potim}")

        if "zbrent" in error:
            incar["IBRION"] = 1
            corrections.append("switched IBRION to 1")
            poscar_filename = os.path.join(dir, "POSCAR")
            contcar_filename = os.path.join(dir, "CONTCAR")
            shutil.copyfile(contcar_filename, poscar_filename)
            corrections.append("copied the CONTCAR over to the POSCAR")

        if "too_few_bands" in self.errors:
            # Grab the current number of bands. First check the INCAR and if
            # it isn't there, jump to the OUTCAR.
            if "NBANDS" in incar:
                nbands_current = incar["NBANDS"]
            else:
                outcar_filename = os.path.join(dir, "OUTCAR")
                with open(outcar_filename) as file:
                    lines = file.readlines()
                # Go through the lines until we find the NBANDS one. The value
                # should be at the very end of the line.
                for line in lines:
                    if "NBANDS" in line:
                        nbands_current = int(line.split()[-1])
                        break
            # increase the number of bands by 10%
            nbands_new = nbands_current * 1.1
            incar["NBANDS"] = nbands_new
            corrections.append(
                f"switch NBANDS from {nbands_current} to {nbands_new} (+10%)"
            )

        if "pssyevx" in error:
            incar["ALGO"] = "Normal"
            corrections.append("switched ALGO to Normal")

        if "eddrmm" in error:
            # RMM algorithm is not stable for this calculation
            if incar.get("ALGO", "Normal") in ["Fast", "VeryFast"]:
                incar["ALGO"] = "Normal"
                corrections.append("switched ALGO to Normal")
            else:
                # Halve the POTIM
                current_potim = incar.get("POTIM", 0.5)
                new_potim = current_potim / 2
                incar["POTIM"] = new_potim
                corrections.append(f"switch POTIM from {current_potim} to {new_potim}")

            # Check the current ICHARG setting, where default is 0
            # If the ICHARG is less than 10, then we want to delete the CHGCAR
            # and WAVECAR to ensure the next run is a clean start.
            current_icharg = incar.get("ICHARG", 0)
            if current_icharg < 10:
                os.remove(os.path.join(dir, "CHGCAR"))
                os.remove(os.path.join(dir, "WAVECAR"))
            corrections.append("deleted CHGCAR and WAVECAR")

        if "edddav" in error:
            current_icharg = incar.get("ICHARG", 0)
            if current_icharg < 10:
                os.remove(os.path.join(dir, "CHGCAR"))
                corrections.append("deleted CHGCAR")
            incar["ALGO"] = "All"
            corrections.append("switched ALGO to Normal")

        if "grad_not_orth" in error:
            if incar.get("ISMEAR", 1) < 0:
                incar["ISMEAR"] = "0"
                corrections.append("switched ALGO to Normal")
                incar["SIGMA"] = 0.05
                corrections.append("switched SIGMA to 0.05")

        if "zheev" in error:
            if incar.get("ALGO", "Fast") != "Exact":
                incar["ALGO"] = "Exact"
                corrections.append("switched Algo to Exact")

        if "elf_kpar" in error:
            incar["KPAR"] = 1
            corrections.append("switched KPAR to 1")

        if "rhosyg" in error:
            if incar.get("SYMPREC", 1e-4) == 1e-4:
                incar["ISYM"] = 0
                corrections.append("switched KPAR to 1")
            incar["SYMPREC"] = 1e-4
            corrections.append("switched SYMPREC to 1e-4")

        if "posmap" in error:
            # VASP advises to decrease or increase SYMPREC by an order of magnitude
            # the default SYMPREC value is 1e-5

            error_counts["posmap"] = error_counts.get("posmap", 0) + 1

            if error_counts["posmap"] == 0:
                # first, reduce by 10x
                current_symprec = incar.get("SYMPREC", 1e-5)
                new_symprec = current_symprec / 10
                incar["SYMPREC"] = new_symprec
                corrections.append(
                    f"switched SYMPREC from {current_symprec} to {new_symprec}"
                )

            elif error_counts["posmap"] == 1:
                # next, increase by 100x (10x the original because we descreased
                # by 10x in the first try.)
                current_symprec = incar.get("SYMPREC", 1e-5)
                new_symprec = current_symprec * 100
                incar["SYMPREC"] = new_symprec
                corrections.append(
                    f"switched SYMPREC from {current_symprec} to {new_symprec}"
                )

        if "point_group" in error:
            incar["ISYM"] = 0
            corrections.append("switched ISYM to 0")

        if "symprec_noise" in error:
            if incar.get("ISYM", 2) > 0 and incar.get("SYMPREC", 1e-5) > 1e-6:
                incar["SYMPREC"] = 1e-6
                corrections.append("switched SYMPREC to 1e-6")
            else:
                incar["ISYM"] = 0
                corrections.append("switched ISYM to 0")

        # Now that we made all of our changes to the INCAR, we can rewrite it
        incar.to_file(incar_filename)

        #!!! BUG: rewrite simmate_error_counts!

        # return the list of corrections we made
        return corrections



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.relaxation.mit import MITRelaxation

# This class used pymatgen's MITMDSet as it basis for settings.


class MITDynamicsTask(MITRelaxation):
    """
    Runs a molecular dynamics simulation using MIT Project settings.

    This will run from 300-1200 Kelvin over 10,000 steps of 2 femtoseconds, and
    it will keep the lattice+volume fixed.

    Provide your structure as the desired supercell, as the setup of this
    calculation does not modify your input structure.
    """

    # The settings used for this calculation are based on the MITRelaxation, but
    # we are updating/adding new settings here.
    # !!! we hardcode temperatures and time steps here, but may take these as inputs
    # in the future
    incar = MITRelaxation.incar.copy()
    incar.update(
        dict(
            TEBEG=300,  # start temperature
            TEEND=1200,  # end temperature
            NSW=10000,  # number of steps
            EDIFF__per_atom=1e-06,
            LSCALU=False,
            LCHARG=False,
            LPLANE=False,
            LWAVE=True,
            ISMEAR=0,
            NELMIN=4,
            LREAL=True,
            BMIX=1,
            MAXMIX=20,
            NELM=500,
            NSIM=4,  # number of bands to run in parallel  # !!! Should this be moved to config?
            ISYM=0,  # turn off symmetry
            ISIF=0,  # only update atom sites
            IBRION=0,  # turns on molecular dynamics
            NBLOCK=1,
            KBLOCK=100,
            SMASS=0,
            POTIM=2,  # time step (in fs)
            PREC="Low",
        )
    )
    # because we no longer use LDAU, we can also remove all relevent settings from
    # the incar for clarity.
    incar.pop("multiple_keywords__smart_ldau")
    # Likewise, we set ISMEAR=0 and EDIFF above, so we no longer need smart_ismear
    incar.pop("multiple_keywords__smart_ismear")
    incar.pop("EDIFF")

    # For now, I turn off all error handlers.
    # TODO
    error_handlers = []



# -*- coding: utf-8 -*-

import os

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class RotationMatrix(ErrorHandler):
    """
    This a simple error handler that is active when VASP struggles to find the
    rotation matrix. VASP gives us the suggested fix directly, which is to
    simply increase the symmetry precision.
    """

    # run this while the VASP calculation is still going
    is_monitor = True

    # we assume that we are checking the vasp.out file
    filename_to_check = "vasp.out"

    # These are the error messages that we are looking for in the file
    possible_error_messages = ["rotation matrix was not found (increase SYMPREC)"]

    def correct(self, directory):

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # increase the precision
        incar["SYMPREC"] = 1e-8

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return "switched SYMPREC to 1e-8"



# -*- coding: utf-8 -*-

from django import forms

from simmate.utilities import get_chemical_subsystems


class ChemicalSystemForm(forms.Form):

    # This is the most basic form for querying data, which only takes
    # the chemical system. This way users can quickly query with a composition
    # such as "Y-C-F". We also ask if they want the subsystem too. For example,
    # Y-C-F would also include Y, C, F, Y-C, Y-F, etc. in the search results.

    # We want a lot of control here and this form will be applied to multiple
    # tables/models, so we don't use django-filter here.

    # Searching by Chemical System
    include_subsystems = forms.BooleanField(label="Include Subsytems", required=False)
    chemical_system = forms.CharField(label="Chemical System", max_length=20)

    # Searching by Database Providers
    aflow = forms.BooleanField(required=False)
    cod = forms.BooleanField(required=False)
    jarvis = forms.BooleanField(required=False)
    materials_project = forms.BooleanField(required=False)
    simmate = forms.BooleanField(required=False)
    oqmd = forms.BooleanField(required=False)

    def clean_chemical_system(self):

        # Our database expects the chemical system to be given in alphabetical
        # order, but we don't want users to recieve errors when they search for
        # "Y-C-F" instead of "C-F-Y". Therefore, we fix that for them here! We
        # also check if the user wants the chemical subsystems included.

        # Grab what the user submitted
        chemical_system = self.cleaned_data["chemical_system"]

        # TODO:
        # Make sure that the chemical system is made of valid elements and
        # separated by hyphens

        # check if the user wants subsystems included (This will be True or False)
        if self.cleaned_data["include_subsystems"]:
            systems_cleaned = get_chemical_subsystems(chemical_system)

        # otherwise just clean the single system
        else:
            # Convert the system to a list of elements
            systems_cleaned = chemical_system.split("-")
            # now recombine the list back into alphabetical order
            systems_cleaned = ["-".join(sorted(systems_cleaned))]
            # NOTE: we call this "systems_cleaned" and put it in a list so
            # that our other methods don't have to deal with multiple cases
            # when running a django filter.

        # now return the cleaned value. Note that this is now a list of
        # chemical systems, where all elements are in alphabetical order.
        return systems_cleaned



# -*- coding: utf-8 -*-

"""
This file hosts common functions that are used throughout Simmate
"""

import os
import itertools
from tempfile import TemporaryDirectory, mkdtemp
import shutil
import sys

from typing import List, Union


def get_conda_env() -> str:
    """
    Grab the name of the activate conda environment and returns it as a string.
    If there is no env, then an empty string is returned.
    """
    # Check the list of python paths and grab the first path that has "envs" in it.
    # Assume we don't have a conda env until proven otherwise
    env_name = ""
    for path in sys.path:
        if "envs" in path:
            # split the path into individual folder names (os.sep gives / or \\)
            folders = path.split(os.sep)
            # the conda env name will the name immediately after the /envs.
            # example path is '/home/jacksund/anaconda3/envs/simmate_dev/lib/python3.9'
            # where we want the name simmate_dev here.
            env_name = folders[folders.index("envs") + 1]
            # once we have found this, we can exit the loop
            break

    return env_name


def get_doc_from_readme(file: str) -> str:
    """
    Loads the docstring from a README.md file in the same directory.

    This is commonly used in __init__.py files because we like having our
    documentation isolated (so that github renders it).

    To use, simply pass the file property:

    ``` python
    from simmate.utilities import get_doc_from_readme

    __doc__ = get_doc_from_readme(__file__)
    ```

    This is an alternative to using "include" in rst files, which
    [pdoc recommends](https://pdoc.dev/docs/pdoc.html#include-markdown-files).
    We prefer this utility because it allows Spyder to load the docs -- although
    it's slower in production (bc of opening/closing files).
    """

    # We assume the file is in the same directory and named "README.rst"
    file_directory = os.path.dirname(os.path.abspath(file))
    with open(
        os.path.join(file_directory, "README.md"),
        encoding="utf-8",
    ) as doc_file:
        doc = doc_file.read()
    return doc


def get_directory(directory: Union[str, TemporaryDirectory] = None) -> str:
    """
    Initializes a directory.

    There are many cases where the user can choose their working directory
    for a calculation, and they may want to provide their directory in various
    formats. This includes... None, a string, or a TemporaryDirectory instance.
    Based on the input, this function does the following:
      - `None`:
          returns the full path to a new folder inside python's
          current working directory named "simmate-task-<randomID>"
      - `TemporaryDirectory`:
          returns the full path to the given temp directory
      - `str`:
          makes the directory if it doesnt exist and then returns the path

    Parameters
    ----------
    - `directory`:
        Either None, a path to the directory, or a tempdir. The default is None.

    Returns
    -------
    - `directory`:
        The path to the initialized directory as a string
    """

    # if no directory was provided, we create a new folder within the current
    # working directory. All of these folders are named randomly.
    # Note: we can't name these nicely (like simmate-task-001) because that will
    # introduce race conditions when making these folders in production.
    if not directory:

        # create a directory in the current working directory. Note, even though
        # we are creating a "TemporaryDirectory" here, this directory is never
        # actually deleted.
        directory = mkdtemp(prefix="simmate-task-", dir=os.getcwd())
    # if the user provided a tempdir, we want it's name
    elif isinstance(directory, TemporaryDirectory):
        directory = directory.name
    # otherwise make sure the directory the user provided exists and if it does
    # not, then make it!
    else:
        # We use mkdirs instead of mkdir because we want to make these directory
        # recursively. That is, we can do "path/to/newfolder1/newfolder2" where
        # multiple folders can be made with one call.
        # Also if the folder already exists, we don't want to raise an error.
        os.makedirs(directory, exist_ok=True)
    # and return the full path to the directory
    return directory


def make_archive(directory: str):
    """
    Compresses the directory to a zip file of the same name. After compressing,
    it then deletes the original directory.

    Parameters
    ----------
    - `directory`:
        Path to the folder that should be archived
    """
    # This wraps shutil.make_archive to change the default parameters. Normally,
    # it writes the archive in the working directory, but we update it to use the
    # the same directory that the folder being archived. The format is also set
    # to zip
    shutil.make_archive(
        # By default I choose within the current directory and save
        # it as the same name of the directory (+ zip ending)
        base_name=os.path.join(os.path.abspath(directory)),
        # format to use switch to gztar after testing
        format="zip",
        # full path to up tp directory that will be archived
        root_dir=os.path.dirname(directory),
        # directory within root_directory to archive
        base_dir=os.path.basename(directory),
    )
    # now remove the directory we just archived
    shutil.rmtree(directory)


def empty_directory(directory: str, files_to_keep: List[str] = []):
    """
    Deletes all files and folders within a directory, except for those provided
    to the files_to_keep parameter.

    Parameters
    ----------
    - `directory`:
        base directory that should be emptied
    - `files_to_keep`:
        A list of file and folder names within the base directory that should
        not be deleted. The default is [].
    """
    # grab all of the files and folders inside the listed directory
    for filename in os.listdir(directory):
        if filename not in files_to_keep:
            # check if we have a folder or a file.
            # Folders we delete with shutil and files with the os module
            if os.path.isdir(filename):
                shutil.rmtree(filename)  # ignore_errors=False
            else:
                os.remove(filename)


def get_chemical_subsystems(chemical_system: str):
    """
    Given a chemical system, this returns all chemical systems that are also
    contained within it.

    For example, "Y-C" would return ["Y", "C", "C-Y"]. Note that the returned
    list has elements of a given system in alphabetical order (i.e. it gives
    "C-Y" and not "Y-C")

    Parameters
    ----------
    - `chemical_system`:
        A chemical system of elements. Elements must be separated by dashes (-)

    Returns
    -------
    - `subsystems`:
        A list of chemical systems that make up the input chemical system.
    """

    # TODO: this will may be better located elsewhere. Maybe even as a method for
    # the Composition class.

    # Convert the system to a list of elements
    system_cleaned = chemical_system.split("-")

    # Now generate all unique combinations of these elements. Because we also
    # want combinations of different sizes (nelements = 1, 2, ... N), then we
    # put this in a for-loop.
    subsystems = []
    for i in range(len(system_cleaned)):
        # i is the size of combination we want. We now ask for each unique combo
        # of elements at this given size.
        for combo in itertools.combinations(system_cleaned, i + 1):
            # Combo will be a tuple of elements that we then convert back to a
            # chemical system. We also sort this alphabetically.
            #   ex: ("Y", "C", "F") ---> "C-F-Y"
            subsystem = "-".join(sorted(combo))
            subsystems.append(subsystem)
    return subsystems



# -*- coding: utf-8 -*-

from django.urls import path

from simmate.website.local_calculations import views

urlpatterns = [
    #
    # Lists off the different types of calculations
    path(
        route="",
        view=views.all_local_calculations,
        name="local_calculations",
    ),
    #
    # Lists off the different types of relaxations
    path(
        route="relaxations/",
        view=views.relaxations,
        name="relaxations",
    ),
    #
    # TODO: For now I list all calculation types here, but it will be useful
    # to have subapps that do this. I also want to incorporate dynamic loading
    # of the different options.
    path(
        route="relaxations/mit/",
        view=views.mit_about,
        name="mit_about",
    ),
    path(
        route="relaxations/mit/submit",
        view=views.mit_submit,
        name="mit_submit",
    ),
    path(
        route="relaxations/mit/results",
        view=views.mit_all,
        name="mit_all",
    ),
    path(
        route="relaxations/mit/results/<int:mitrelax_id>",
        view=views.mit_single,
        name="mit_single",
    ),
]



# -*- coding: utf-8 -*-

# There are not any custom inputs for Bader analysis. The inputs required
# are actually outputs from VASP or some other DFT code. We show what
# some of those are here:

# TODO these files don't exist yet
# from simmate.calculators.vasp.outputs import CHGCAR, AECCAR0, AECCAR2



# -*- coding: utf-8 -*-

import os

from pymatgen.io.vasp.inputs import Kpoints
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.symmetry.bandstructure import HighSymmKpath
from pymatgen.electronic_structure.plotter import BSPlotter

from simmate.calculators.vasp.inputs import Incar, Poscar, Potcar
from simmate.calculators.vasp.tasks.energy.materials_project import (
    MatProjStaticEnergy,
)


class MatProjBandStructure(MatProjStaticEnergy):
    """
    Calculates the band structure using Materials Project settings.

    Your stucture will be converted to the standardized-primitive unitcell so
    that the high-symmetry K-point path can be used.

    This is also a non self-consistent field (non-SCF) calculation and thus uses
    the a fixed charge density from a previous static energy calculation.

    We do NOT recommend running this calculation on its own. Instead, you should
    use the full workflow, which handles the preliminary relaxation and SCF
    energy calculation for you. This S3Task is only the final step of that workflow.

    See `vasp.workflows.band_structure`.
    """

    # The default settings to use for this calculation.
    # The key thing for band structures is that this follows a static energy
    # calculation. We use the CHGCAR from that previous calculation. We also
    # hold the charge density static during this calculation so this is a
    # non-selfconsistent (non-SCF) calculation.
    incar = MatProjStaticEnergy.incar.copy()
    incar.update(
        ICHARGE=11,
        ISMEAR=0,
        SIGMA=0.01,
    )

    # set the KptGrid or KptPath object
    # TODO: in the future, all functionality of this class will be available
    # by giving a KptPath class here.
    kpoints = None
    kpoints_line_density = 20

    def setup(self, structure, directory):
        """
        Writes input files for this calculation. This differs from the normal
        VaspTask setup because it converts the structure to the standard primative
        first and then writes a KPOINT file with using a highsym path.
        """

        # If requested, we convert to the LLL-reduced unit cell, which aims to
        # be as cubic as possible.
        if self.pre_sanitize_structure:
            structure = structure.copy(sanitize=True)

        # For band structures, we need to make sure the structure is in the
        # standardized primitive form. Note we assume that we are using the
        # "interanational_monoclinc" in pymatgen.
        # We use the same SYMPREC from the INCAR, which is 1e-5 if not set.
        sym_prec = self.incar.get("SYMPREC", 1e-5) if self.incar else 1e-5
        sym_finder = SpacegroupAnalyzer(structure, symprec=sym_prec)
        structure_standardized = sym_finder.get_primitive_standard_structure()
        # check for pymatgen bugs here
        check_for_standardization_bugs(structure, structure_standardized)

        # write the poscar file
        Poscar.to_file(structure, os.path.join(directory, "POSCAR"))

        # Combine our base incar settings with those of our parallelization settings
        # and then write the incar file
        incar = Incar(**self.incar) + Incar(**self.incar_parallel_settings)
        incar.to_file(
            filename=os.path.join(directory, "INCAR"),
            structure=structure,
        )

        ##############
        # we need to find the high-symmetry Kpt path. Note that all of this
        # functionality will be moved to the KptPath class and then extended to
        # vasp.inputs.kpoints class. Until those classes are ready, we just use
        # pymatgen here.
        kpath = HighSymmKpath(structure_standardized)
        frac_k_points, k_points_labels = kpath.get_kpoints(
            line_density=self.kpoints_line_density,
            coords_are_cartesian=False,
        )
        kpoints = Kpoints(
            comment="Non SCF run along symmetry lines",
            style=Kpoints.supported_modes.Reciprocal,
            num_kpts=len(frac_k_points),
            kpts=frac_k_points,
            labels=k_points_labels,
            kpts_weights=[1] * len(frac_k_points),
        )
        kpoints.write_file(os.path.join(directory, "KPOINTS"))
        ##############

        # write the POTCAR file
        Potcar.to_file_from_type(
            structure.composition.elements,
            self.functional,
            os.path.join(directory, "POTCAR"),
            self.potcar_mappings,
        )

    def _write_output_summary(self, directory, vasprun):
        """
        In addition to writing the normal VASP output summary, this also plots
        the bandstructure to "band_structure.png"
        """

        # run the normal out put
        super()._write_output_summary(directory, vasprun)

        bs_plotter = BSPlotter(vasprun.get_band_structure(line_mode=True))
        plot = bs_plotter.get_plot()
        plot_filename = os.path.join(directory, "band_structure.png")
        plot.savefig(plot_filename)


def check_for_standardization_bugs(structure_original, structure_new):

    # In pymatgen, they include this code with the standardization of their
    # structures because there were several bugs in the past and they want to
    # double-check themselves. I'm still using their code to standardize
    # my structures, so I should make this check too.

    vpa_old = structure_original.volume / structure_original.num_sites
    vpa_new = structure_new.volume / structure_new.num_sites

    if abs(vpa_old - vpa_new) / vpa_old > 0.02:
        raise ValueError(
            "Standardizing failed! Volume-per-atom changed... "
            f"old: {vpa_old}, new: {vpa_new}"
        )

    sm = StructureMatcher()
    if not sm.fit(structure_original, structure_new):
        raise ValueError("Standardizing failed! Old structure doesn't match new.")



# -*- coding: utf-8 -*-

from django.urls import include, path

from rest_framework import routers

from simmate.website.rest_api.views import (
    SpacegroupViewSet,
    MatProjViewSet,
    JarvisViewSet,
    AflowViewSet,
    OqmdViewSet,
    CodViewSet,
)


router = routers.DefaultRouter()
router.register("space-groups", SpacegroupViewSet)
router.register("materials-project", MatProjViewSet)
router.register("jarvis", JarvisViewSet)
router.register("aflow", AflowViewSet)
router.register("oqmd", OqmdViewSet)
router.register("cod", CodViewSet)

urlpatterns = [
    # Our router automatically handle each possible endpoint for us! We stick
    # with the default REST templates and views too.
    path(
        route="",
        view=include(router.urls),
        # name="api-root", # This is set by default
        # To view all other urls that this maps, you can use django_extensions
        # to list all of them. The command is... python manage.py show_urls
    ),
    #
    # Additionally, we include login URLs. This doesn't do anything other than
    # let the user login on the default REST interface.
    path(
        route="accounts/",
        view=include("rest_framework.urls"),
        name="rest_login",
    ),
]



# -*- coding: utf-8 -*-

import os

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class RotationMatrix(ErrorHandler):
    """
    This a simple error handler that is active when VASP finds an issue with the
    rotation matrix.
    """

    # run this while the VASP calculation is still going
    is_monitor = True

    # we assume that we are checking the vasp.out file
    filename_to_check = "vasp.out"

    # These are the error messages that we are looking for in the file
    possible_error_messages = [
        "Found some non-integer element in rotation matrix",
        "SGRCON",
    ]

    def correct(self, directory):

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # switch to gamma-centered mesh
        incar["KGAMMA"] = True
        correction = "switched KGAMMA to True"
        # TODO: 2nd attempt is to turn ISYM=0

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return correction



# -*- coding: utf-8 -*-

from django.contrib import admin
from django.urls import include, path

from simmate.website.core import views

urlpatterns = [
    #
    # This is the path to the homepage (just simmate.org)
    path(route="", view=views.home, name="home"),
    #
    # This is the built-in admin site that django provides
    path(route="admin/", view=admin.site.urls, name="admin"),
    #
    # This is profile system with login/logout
    path(
        route="accounts/",
        view=include("simmate.website.accounts.urls"),
        name="accounts",
    ),
    #
    #
    path(route="extras/", view=views.extras, name="extras"),
    #
    #
    path(
        route="third-parties/",
        view=include("simmate.website.third_parties.urls"),
        name="third_parties",
    ),
    #
    #
    # This maps to our REST api endpoints
    path(
        route="rest-api/",
        view=include("simmate.website.rest_api.urls"),
        name="rest_api",
    ),
    # All local calculations are stored at this endpoint
    path(
        route="local-calculations/",
        view=include("simmate.website.local_calculations.urls"),
        name="local_calculations",
    ),
    #
    # This app is for viewing crystal structures in a 3D viewport
    path(
        route="structure-viewer/",
        view=include("simmate.website.structure_viewer.urls"),
        name="structure_viewer",
    ),
]



# -*- coding: utf-8 -*-

from django.urls import path

from simmate.website.structure_viewer import views

urlpatterns = [
    path(
        route="",
        view=views.structure_viewer,
        name="test",
    ),
]



# -*- coding: utf-8 -*-

from simmate.workflow_engine.utilities import s3task_to_workflow
from simmate.calculators.vasp.tasks.relaxation import (
    MatProjRelaxation as MPRelaxationTask,
)
from simmate.calculators.vasp.database.relaxation import (
    MatProjRelaxation as MPRelaxationResults,
)

workflow = s3task_to_workflow(
    name="Materials Project Relaxation",
    module=__name__,
    project_name="Simmate-Relaxation",
    s3task=MPRelaxationTask,
    calculation_table=MPRelaxationResults,
    register_kwargs=["prefect_flow_run_id", "structure", "source"],
)



# -*- coding: utf-8 -*-

import os
from pathlib import Path

import yaml

from pymatgen.io.vasp.outputs import Vasprun

from simmate.calculators.vasp.inputs import Incar, Poscar, Kpoints, Potcar
from simmate.workflow_engine import S3Task


def get_default_parallel_settings():
    # We load the user's default parallel settings from ~/simmate/vasp/INCAR_parallel_settings
    # If this file doesn't exist, then we just use an empty dictionary
    settings_filename = os.path.join(
        Path.home(), "simmate", "vasp", "INCAR_parallel_settings"
    )
    if os.path.exists(settings_filename):
        return Incar.from_file(settings_filename)
    else:
        return {}


class VaspTask(S3Task):

    # Vasp calculations always need an input structure
    requires_structure = True

    # BUG:
    # Prefect is unfortunately unable access a task result's attribute when
    # building a flow. So I am unable to do things like...
    #
    #   with Workflow("example") as workflow:
    #       output = example_task()
    #       other_example_task(structure=output.structure)
    #
    # instead I need to make sure my output is a dictionary like so...
    #
    #   with Workflow("example") as workflow:
    #       output = example_task()
    #       other_example_task(structure=output["structure"])
    #
    # This controls whether we return just the result or a dict of result and
    # final structure. Really, the result (a Vasprun object) contains the final
    # structure via result.final_structure, BUT Prefect is causing problems here.
    # For now, I only set this to True during relaxations.
    return_final_structure = False

    # The command to call vasp in the current directory
    # TODO: add support for grabbing a user-set default from their configuration
    command = "vasp_std > vasp.out"

    # set the default vasp settings from a dictionary. This is the one thing
    # you *must* set when subclassing VaspTask. An example is:
    #   incar = dict(NSW=0, PREC="Accurate", KSPACING=0.5)
    incar = None

    # We also load any parallel settings to add on to the base incar. These
    # should not effect the calculation in any way, but they are still selected
    # based on the computer specs and what runs fastest on it. Therefore, these
    # settings are loaded from ~/simmate/vasp/INCAR_parallel_settings by default.
    # This can also be overwritten as well.
    incar_parallel_settings = get_default_parallel_settings()

    # TODO: add options for poscar formation
    # add_selective_dynamics=False
    # add_velocities=False
    # significant_figures=6

    # set the KptGrid or KptPath object
    # TODO - KptGrid is just a float for now
    # NOTE - this is optional because you can have KSPACING as an INCAR argument.
    # If KSPACING is set above, we ignore whatever is set here.
    kpoints = None

    # This directs which Potcar files to grab. You would set this to a string
    # of what you want, such as "PBE", "PBE_GW", or "LDA"
    functional = None

    # This is an optional parameter to override Simmate's default selection of
    # potentials based off of the functional chosen. The defaults are located
    # in simmate.calculators.vasp.inputs.potcar_mappings. You can supply your
    # own mapping dictionary or you can take ours and update the specific
    # potentials you'd like. For example:
    #   from simmate.calculators.vasp.inputs.potcar_mappings import PBE_ELEMENT_MAPPINGS
    #   element_mappings = PBE_ELEMENT_MAPPINGS.copy().update({"C": "C_h"})
    # or if you only use Carbon and don't care about other elements...
    #   element_mappings = {"C": "C_h"}
    # Read more on this inside the Potcar class and be careful with updating!
    potcar_mappings = None

    # In somecases we still want results from calculations that did NOT converge
    # successfully. This flag controls whether or not we raise an error when
    # the calculation failed to converge.
    # OPTIMIZE: What if I updated the ErrorHandler class to allow for "warnings"
    # instead of raising the error and applying the correction...? This functionality
    # could then be moved to the UnconvergedErrorHandler. I'd have a fix_error=True
    # attribute that is used in the .check() method. and If fix_error=False, I
    # simply print a warning & also add that warning to simmate_corrections.csv
    confirm_convergence = True

    # In some cases, we may want to sanitize the structure during our setup().
    # This means converting to the LLL-reduced primitive cell.
    pre_sanitize_structure = False
    # The same might be true for converting the structure to the standard primitive
    # of a structure. For example, this is required when calculating band structures.
    pre_standardize_structure = False  # TODO: not implemented yet

    def __init__(
        self,
        incar=None,
        kpoints=None,
        functional=None,
        potcar_mappings=None,
        confirm_convergence=None,
        pre_sanitize_structure=None,
        # To support other options from the Simmate SSSTask and Prefect Task
        **kwargs,
    ):

        # if any of these input parameters were given, overwrite the default
        # Note to python devs: this odd formatting is because we set our defaults
        # to None in __init__ while our actual default values are define above
        # as class attributes. This may seem funky at first glance, but it
        # makes inheriting from this class extremely pretty!
        # This code is effectively the same as @defaults_from_attrs(...)
        if incar:
            self.incar = incar
        if kpoints:
            self.kpoints = kpoints
        if functional:
            self.functional = functional
        if potcar_mappings:
            self.potcar_mappings = potcar_mappings
        if confirm_convergence:
            self.confirm_convergence = confirm_convergence
        if pre_sanitize_structure:
            self.pre_sanitize_structure = pre_sanitize_structure

        # now inherit from parent SSSTask class
        super().__init__(**kwargs)

    def setup(self, structure, directory):

        # If requested, we convert to the LLL-reduced unit cell, which aims to
        # be as cubic as possible.
        if self.pre_sanitize_structure:
            structure = structure.copy(sanitize=True)

        # write the poscar file
        Poscar.to_file(structure, os.path.join(directory, "POSCAR"))

        # Combine our base incar settings with those of our parallelization settings
        # and then write the incar file
        incar = Incar(**self.incar) + Incar(**self.incar_parallel_settings)
        incar.to_file(
            filename=os.path.join(directory, "INCAR"),
            structure=structure,
        )

        # if KSPACING is not provided in the incar AND kpoints is attached to this
        # class instance, then we write the KPOINTS file
        if self.kpoints and ("KSPACING" not in self.incar):
            Kpoints.to_file(
                structure,
                self.kpoints,
                os.path.join(directory, "KPOINTS"),
            )

        # write the POTCAR file
        Potcar.to_file_from_type(
            structure.composition.elements,
            self.functional,
            os.path.join(directory, "POTCAR"),
            self.potcar_mappings,
        )

    def workup(self, directory):
        """
        This is the most basic VASP workup where I simply load the final structure,
        final energy, and (if requested) confirm convergence. I will likely make
        this a common function for this vasp module down the road.
        """

        # load the xml file and all of the vasprun data
        try:
            vasprun = Vasprun(
                filename=os.path.join(directory, "vasprun.xml"),
                exception_on_bad_xml=True,
            )
        except:
            print(
                "XML is malformed. This typically means there's an error with your"
                " calculation that wasn't caught by your ErrorHandlers. We try"
                " salvaging data here though."
            )
            vasprun = Vasprun(
                filename=os.path.join(directory, "vasprun.xml"),
                exception_on_bad_xml=False,
            )
            vasprun.final_structure = vasprun.structures[-1]
        # BUG: This try/except is 100% just for my really rough calculations
        # where I don't use any ErrorHandlers and still want the final structure
        # regarless of what when wrong. In the future, I should consider writing
        # a separate method for those that loads the CONTCAR and moves on.

        # write output files/plots for the user to quickly reference
        self._write_output_summary(directory, vasprun)

        # confirm that the calculation converged (ionicly and electronically)
        if self.confirm_convergence:
            assert vasprun.converged

        # OPTIMIZE: see my comment above on the return_final_structure attribute
        if self.return_final_structure:
            return {"structure_final": vasprun.final_structure, "vasprun": vasprun}

        # return vasprun object
        return vasprun

    def _write_output_summary(self, directory, vasprun):
        """
        This is an EXPERIMENTAL feature.

        This prints a "simmate_summary.yaml" file with key output information.

        This method should not be called directly as it used within workup().
        """
        # OPTIMIZE: Ideally, I could take the vasprun object and run to_json,
        # but this output is extremely difficult to read.

        results = vasprun.as_dict()["output"]

        summary = {
            "structure_final": "The final structure is located in the CONTCAR file",
            "energy_final": float(results.get("final_energy", None)),
            "energy_final_per_atom": float(results.get("final_energy_per_atom", None)),
            "converged_electroinc": vasprun.converged_electronic,
            "converged_ionic": vasprun.converged_ionic,
            "fermi_energy": results.get("efermi", None),
            "valence_band_maximum": results.get("vbm", None),
            "conduction_band_minimum": results.get("vbm", None),
        }

        summary_filename = os.path.join(directory, "simmate_summary.yaml")
        with open(summary_filename, "w") as file:
            content = yaml.dump(summary)
            file.write(content)

    @classmethod
    def get_config(cls):
        """
        Grabs the overall settings from the class. This is useful for printing out
        settings for users to inspect.
        """
        return {
            key: getattr(cls, key)
            for key in [
                "__module__",
                "incar",
                "pre_sanitize_structure",
                "incar",
                "functional",
                "potcar_mappings",
                "confirm_convergence",
            ]
        }



# -*- coding: utf-8 -*-

from django.urls import path

from simmate.website.third_parties import views

app_name = "third_parties"
urlpatterns = [
    # This is the main page for this app.
    path(
        route="",
        view=views.home,
        name="third_parties_home",
    ),
]



# -*- coding: utf-8 -*-

from .tetrahedron_mesh import TetrahedronMesh
from .eddrmm import Eddrmm
from .incorrect_smearing import IncorrectSmearingHandler
from .mesh_symmetry import MeshSymmetryErrorHandler
from .unconverged import UnconvergedErrorHandler
from .nonconverging import NonConvergingErrorHandler
from .potim import PotimErrorHandler
from .positive_energy import PositiveEnergyErrorHandler
from .frozen import FrozenErrorHandler
from .large_sigma import LargeSigmaErrorHandler



# -*- coding: utf-8 -*-

from django.shortcuts import render


def structure_viewer(request):
    context = {}
    template = "structure_viewer/main.html"
    return render(request, template, context)



# -*- coding: utf-8 -*-

import os
import time

from fabric import Connection

# BUG: This class is currently broken because Spyder needs to use a jump-host
# in order to connect with the batch-job process.
# ssh -J WarrenLab@warwulf.net WarrenLab@c18
# see https://github.com/spyder-ide/spyder-kernels/issues/359


class SSHConnection(Connection):
    """
    This is a utility for running python code through a remote ssh terminal and
    a PBS/SLURM job queue. It was designed specifically to make Spyder's remote
    kernal connection easier for beginners.

    Read more here:
        http://docs.spyder-ide.org/current/panes/ipythonconsole.html?highlight=ssh#connect-to-a-remote-kernel


    Usage Guide
    ------------

    This class's use is comprised of the following steps:

        1. set up your ssh host, username, and password settings
        2. call setup_remote_kernal and wait for resources to be granted
        3. start a new terminal in Spyder with the connection json file supplied

    Steps 1-2 are completed with the following code, which includes example
    SSH connection settings and SLURM submission command (so be sure to update these):

    .. code-block:: python
        from simmate.configuration.ssh import SSHConnection

        # Example here uses SLURM (sbatch) to submit a job limted to 1 hour.
        # We HIGHLY recommend setting a time limit below 3 hrs to avoid hogging
        # resources outside of your interactive session.
        # This command below is the same as submitting a script with...
        #
        # #SBATCH --nodes=1
        # #SBATCH --ntasks=2
        # #SBATCH --cpus-per-task 1
        # #SBATCH --mem=4GB
        # #SBATCH --partition=p1
        # #SBATCH --time=03:00:00
        #

        # set up the connection
        connection = SSHConnection(
            host="warwulf.net",
            user="WarrenLab",
            connect_kwargs={"password": "xxxxxx"},
            conda_env="test",
            working_directory="/media/synology/user/jack/debug",  # no ending slash!
            submit_command="sbatch -N 1 -n 2 -c 1 --mem=4GB -p p1 -t 03:00:00",
        )

        # Submit your SLURM job, which creates the remote python kernal for you.
        # It will also wait for the job to start, and copy the connection file
        # over to your computer
        connection.setup_remote_kernel()

    Make note of the connection file that is printed out. Once you have this,
    you're all done with this terminal and can close it!

    You can then start a new terminal by selecting "Connect to an existing kernal"
    in Spyder and use the file that is printed out above. This will connect to the
    remote process that you started. To confirm everything is set up and working,
    run the following in your new terminal:

    .. code-block:: python

        # makes sure you're running code remotely. The output should be the
        # working directory of your remote computer
        import os
        os.getcwd()

        # makes sure you have simmate installed on your remote computer
        # if this fails, you need to make sure an environment with Simmate install
        # on your remote computer in addition to your local computer.
        import simmate

    One important thing to remember is that your submitted job will NOT be
    canceled automatically when you close Spyder or your python terminal. Be
    sure to cancel the job when you're done. This is done by typing "exit" in
    your new python terminal.

    """

    def __init__(
        self,
        conda_env: str,
        submit_command: str,
        working_directory: str = ".",
        port: int = 22,  # indicates we are using SSH
        **kwargs,
    ):
        """
        Creates a fabric.Connection where a remote python kernal can be easily
        set up and connected to through SLURM.

        Parameters
        ----------
        conda_env : str
            The conda environment to start with python kernal with. Note, this
            environment is only used within start_remote_kernel() and will not
            be used by the run() method.
        submit_command : str
            The command used to submit to a SLURM or PBS queue system. This should
            be something like "sbatch -N 1 -n 2 -c 2 -p p1 -t 03:00:00" for SLURM.
            PBS is also supported (although not tested in production).
        working_directory : str, optional
            The directory to start the python kernal in. Note, this
            directory is only used within start_remote_kernel() and will not
            be used by the run() method. The default is "." which is the home
            directory.
        port: int
            the remote port. Defaults to 22, which is an SSH connection.
        **kwargs :
            Any arguments passed to fabric.Connection.
        """

        # inherit from parent Connection class
        super().__init__(port=port, **kwargs)

        # and add the extra attributes
        self.conda_env = conda_env
        self.working_directory = working_directory
        self.submit_command = submit_command

        # Warn the user about how PBS has not been tested
        if submit_command.startswith("qsub"):
            raise Exception(
                "This class has not been tested with PBS. If you have a cluster "
                "that we can test with, please reach out to our team and we can "
                "can implement this feature!"
            )

        # We assume the output file for SLURM so stop the user if they tried
        # overwriting this.
        if submit_command.startswith("sbatch") and any(
            x in submit_command for x in ["-o ", "--output "]
        ):
            raise Exception(
                "For this class to work properly, please do not specify the output "
                "filename. If this is a problem for your team, let our team know!"
            )

    def start_remote_kernel(self, sleep_step: int = 3):
        """
        Completes the following steps:
            1. Submit job to cluster which will create a python kernel
            2. Waits for the SLURM/PBS job to start
            3. Loads the connection filename from the batch job output file
            4. Copies the connection file to the local computer

        Parameters
        ----------
        sleep_step : int, optional
            Time to wait between checking if the batch job started. The default
            is 3 seconds.
        """

        # -------------------------------------------------

        # STEP 1: Submit job to cluster

        print("Submitting job to remote cluster...")

        # switch into the working directory
        with self.cd(self.working_directory):
            # and establish the correct conda env before running any command
            with self.prefix(f"source activate {self.conda_env}"):

                # This command starts a python kernel that runs endlessly
                kernel_command = "python -m spyder_kernels.console"

                # How we submit this kernel_command to SLURM/PBS depends on which
                # queue system we are using.
                # SLURM uses "sbatch ... --wrap ..."
                if self.submit_command.startswith("sbatch"):
                    output = self.run(
                        f"{self.submit_command} --wrap '{kernel_command}'"
                    )
                    # The job id is the at the end of the line for output
                    job_id = output.stdout.strip().split()[-1]

                # PBS uses "qsub ... -- ..."
                # elif self.submit_command.startswith("qsub"):
                #     self.run(f"{self.submit_command} -- {kernel_command}")
                # Other systems aren't supported yet and should raise an error.
                else:
                    raise Exception(
                        "The submit_command you provided doesn't match SLURM or "
                        " PBS foramts. SLURM should use sbatch, and PBS should use qsub."
                    )

        # -------------------------------------------------

        # STEP 2: wait for the SLURM/PBS job to start

        print("Waiting for batch job to start...")

        # the slurm output will be named something like /path/to/dir/slurm-12345.out
        # We can't use os.path.join here because we may be running this on Windows
        # when really want os.path.join to act like the remote computer.
        # BUG: for now, we assume the remote computer is Linux.
        if self.submit_command.startswith("sbatch"):
            output_filename = self.working_directory + f"/slurm-{job_id}.out"
        else:
            raise Exception(
                "The submit_command you provided doesn't match SLURM or "
                " PBS foramts. SLURM should use sbatch, and PBS should use qsub."
            )

        # We simply wait until an output file shows up, and once it does, we know
        # that the job has started.
        print(f"Watching for file to exist: {output_filename}")
        while True:
            # check if the output file was written yet
            if not self.run(f"test -f {output_filename}", warn=True).failed:
                break
            # if not, sleep for X seconds until checking again
            time.sleep(sleep_step)

        # Once we know the file exists, we still want to wait 3 extra seconds to
        # give our kernel time to setup. This avoids a race condition.
        time.sleep(3)

        print("Job has started.")

        # -------------------------------------------------

        # STEP 3: Load the connection filename from the batch job output file

        print("Loading connection file location...")

        # transfer the file to the local desktop and grab the filename
        output_filename_local = self.get(output_filename).local

        # Read the file contents for the json connection filename, which will
        # be at the very end of the file.
        with open(output_filename_local) as output_file:
            lines = output_file.readlines()
            connection_filename = lines[-1].split()[-1]
        # we are done with this file so we can delete it
        os.remove(output_filename_local)

        # and establish the correct conda env before running any command
        with self.prefix(f"source activate {self.conda_env}"):
            # grab the directory which holds all kernel json files
            directory = self.run("jupyter --runtime-dir").stdout.strip()
        # We can't use os.path.join here because we may be running this on Windows
        # when really want os.path.join to act like the remote computer.
        # BUG: for now, we assume the remote computer is Linux.
        connection_filename_full = directory + f"/{connection_filename}"

        # -------------------------------------------------

        # STEP 4: Copy the connection file to the local computer

        print("Transfering connection file to local computer...")

        result = self.get(connection_filename_full)

        print(
            f"\n\nSuccess! You will find your connection file at:\n {result.local}\n\n"
        )
        print(
            "Up next, go to Spyder > Console options (bottom right) > Connect to "
            "an Existing Kernal. You can close this console whenever.\n"
            "NOTE: when you are done with your remote kernal, type 'exit' in your "
            "new console so that the kernel (and batch job) are terminated."
        )



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import table_column, Calculation

from typing import List


class NestedCalculation(Calculation):
    """
    A nested calculation is a workflow made up of other workflows. For example,
    we may want to run a workflow that runs a series of relaxations. Or maybe
    a relaxation, then energy, then bandstrucuture calculation. This table
    is for keeping track of workflows ran in series like this.

    Typically, you'll use the `create_subclass_from_calcs` method to create a
    subclass of this table.
    """

    class Meta:
        abstract = True
        app_label = "local_calculations"

    # @abstractproperty
    # child_calculation_tables = [...]

    # TODO:
    # corrections
    # should this be a list of all modifications? It could maybe be used to
    # carry fixes (such as smearing) accross different calcs.
    # Or maybe a method that just lists the corrections of each subcalc?
    # For now I don't use this though. This line removes the field.
    corrections = None

    @classmethod
    def create_subclass_from_calcs(
        cls,
        name: str,
        child_calculation_tables: List[Calculation],
        module: str,
        **extra_columns,
    ):
        """
        Dynamically creates a subclass of NestedCalculation -- and handles linking
        together all child calculation tables.

        `simmate.calculators.vasp.database.relaxation` shows an example of creating
        a table from this class:

        ``` python
        StagedRelaxation = NestedCalculation.create_subclass_from_calcs(
            "StagedRelaxation",
            [
                Quality00Relaxation,
                Quality01Relaxation,
                Quality02Relaxation,
                Quality03Relaxation,
                Quality04Relaxation,
            ],
            module=__name__,
        )
        ```

        To add custom columns, you can do the following:

        ``` python
        from simmate.database.base_data_types import table_column

        StagedRelaxation = NestedCalculation.create_subclass_from_calcs(
            ... # everything is the same as above
            custom_column_01=table_column.FloatField()
        )
        ```

        Parameters
        ----------
        - `name` :
            Name of the subclass that is output.
        - `child_calculation_tables` :
            list of database tables for the nested workflows. This table links
            these sub-tables together so results can be viewed from each step.
        - `module` :
            name of the module this subclass should be associated with. Typically,
            you should pass __name__ to this.
        **extra_columns : TYPE
            Additional columns to add to the table. The keyword will be the
            column name and the value should match django options
            (e.g. table_column.FloatField())

        Returns
        -------
        NewClass :
            A subclass of NestedCalculation.

        """

        # BUG: I assume a workflow won't point to the save calculation table
        # more than once... What's a scenario where this isn't true?
        # I can only think of mult-structure workflows (like EvolutionarySearch)
        # which I don't give their own table for now.
        new_columns = {}
        for child_calc in child_calculation_tables:
            new_column = table_column.OneToOneField(
                child_calc,
                on_delete=table_column.CASCADE,
                # related_name=...,
                blank=True,
                null=True,
            )
            new_columns[f"{child_calc._meta.model_name}"] = new_column

        # Now put all the fields together to make the new class
        NewClass = cls.create_subclass(
            name,
            **new_columns,
            **extra_columns,
            # also have child calcs list as an attribute
            child_calculation_tables=child_calculation_tables,
            module=module,
        )

        # we now have a new child class and avoided writing some boilerplate code!
        return NewClass

    def update_calculation(self):

        raise Exception(
            "NestedCalculation datatable is experimental so this method doesn't "
            "work yet."
        )

        # BUG: This assumes we ran all calculations within the same directory,
        # which isn't true in all cases.
        for child_calc_table in self.child_calculation_tables:
            if child_calc_table.objects.filter(directory=self.directory).exists():
                child_calc = child_calc_table.objects.get(directory=self.directory)
                setattr(self, child_calc._meta.model_name, child_calc)
        self.save()



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import table_column, Structure


class AflowPrototype(Structure):

    # While this is still a table of structures, this should instead be viewed
    # as a table of structure-types. All of the entries in this table come
    # from the AFLOW Encyclopedia of Crystallographic Prototypes:
    #   http://www.aflowlib.org/prototype-encyclopedia/

    """Base Info"""

    mineral_name = table_column.CharField(max_length=75, blank=True, null=True)
    aflow_id = table_column.CharField(max_length=30)
    pearson_symbol = table_column.CharField(max_length=6)
    strukturbericht = table_column.CharField(max_length=6)
    nsites_wyckoff = table_column.IntegerField()

    """ Model Methods """

    @property
    def name(self):
        """
        This helps piece together the name of the prototype in a user-friendly
        format. We start by checking if there is a mineral_name to use, and we
        also use the prototype's composition.

        An example of a structure with a mineral name is...
            Cinnabar (HgS) Structure-type
        And an example of a structure without a mineral name is..
            CaC6 Structure-type
        """
        if self.mineral_name:
            return f"{self.mineral_name} ({self.formula_reduced}) Structure-type"
        else:
            return f"{self.formula_reduced} Structure-type"

    """ Django App Association """

    class Meta:
        app_label = "prototypes"



# -*- coding: utf-8 -*-

"""
This file exists strictly for when you're testing with Simmate and don't want
to write out long imports every single time. When writing modules, you should
never use these shortcuts because they actually take longer to run. 

So if you're a beginner, use the shortcut. If you're contributing to Simmate
code or really care about speed of your code, avoid the shortcut.

By example, the import below is a shortcut for the longer line below it. They 
both load the same function, but have two key differences: (1) The shortcut is 
obiously easier to remember and write -- so it's best for when you're quickly 
testing things via Spyder; and (2) the shortcut loads slower (typically 1 second)
than the full import. This is because the shortcut module (this file) loads other
convient modules too. So you're actually loading extra things!

``` python    
# the convenient shortcut
from simmate.shortcuts import setup

# the faster import that does the same thing
from simmate.configuration.django import setup_full
```

"""

# This import sets up Django so we can connect with the Simmate database
from simmate.configuration.django import setup_full as setup



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import table_column, Structure


class CodStructure(Structure):
    """
    Crystal structures from the [COD](http://www.crystallography.net/cod/) database.

    Currently, this table only stores the strucure, plus comments on whether the
    sturcture is ordered or has implicit hydrogens.
    """

    # Make sure Django knows which app this is associated with
    class Meta:
        app_label = "third_parties"

    base_info = ["id", "structure_string", "energy_above_hull"]
    """
    The base information for this database table. All other columns can be calculated
    using the columns in this list.
    """

    id = table_column.CharField(max_length=25, primary_key=True)
    """
    The id used to represent the structure (ex: "jvasp-12345")
    """

    is_ordered = table_column.BooleanField()
    """
    whether the structure contains disordered sites (i.e. mixed occupancies)
    """

    has_implicit_hydrogens = table_column.BooleanField()
    """
    whether the structure has implicit Hydrogens. This means there should be
    Hydrogens in the structure, but they weren't explicitly drawn.
    """

    source = "COD"
    """
    Where this structure and data came from.
    """

    source_doi = "https://doi.org/10.1107/S0021889809016690"
    """
    Source paper that must be referenced if this dataset is used.
    """

    # These fields overwrite the default Structure fields due to a bug.
    chemical_system = table_column.TextField()
    formula_full = table_column.TextField()
    formula_reduced = table_column.TextField()
    formula_anonymous = table_column.TextField()
    # BUG: We can't use CharField for the COD database because there are a number
    # of structures that have 20+ elements in them and are disordered. The disordered
    # aspect throws issues in pymatgen where formulas can be returned as long
    # floats (ex: Ca2.1234567N). Until this is fixed and cleaned up, I'll
    # need to use TextField instead of CharField for these fields.

    @property
    def external_link(self) -> str:
        """
        URL to this structure in the COD website.
        """
        # All COD structures have their data mapped to a URL in the same way
        # ex: https://www.ctcms.nist.gov/~knc6/static/JARVIS-DFT/JVASP-1234.xml
        # we store the id as "cod-123" so we need to convert this to uppercase
        id_formatted = self.id.split("-")[-1]
        return f"http://www.crystallography.net/cod/{id_formatted}.html"



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import DatabaseTable, table_column


class Calculation(DatabaseTable):
    """
    A calculation is synonmyous with a Prefect "Flow-Run". This entire table together
    can be viewed as the "Flow" which may have extra metadata. This could
    include values such as...
        - workflow_name
        - simmate_version
        - [[thirdparty]]_version (i.e. vasp_version="5.4.4")
    """

    class Meta:
        abstract = True
        app_label = "local_calculations"

    base_info = [
        "directory",
        "prefect_flow_run_id",
        "created_at",
        "updated_at",
        "corrections",
    ]
    """
    The base information for this database table. All other columns can be calculated
    using the columns in this list.
    """

    directory = table_column.CharField(
        max_length=250,
        blank=True,
        null=True,
    )
    """
    This is the folder that the workflow was ran in. It will be something
    like... "/path/to/simmate-task-abc123"
    """
    # TODO: maybe add host computer name too? (via platform.node())

    prefect_flow_run_id = table_column.CharField(
        max_length=50,
        blank=True,
        null=True,
        unique=True,
    )
    """
    If this was submitted through Prefect, there is a lot more metadata available
    for this calculation. Simmate does not store this data directly, but instead
    we store the flow-run-id so that the user may look this up in the Prefect
    database if they wish.
    An example id is... d8a785e1-e344-463a-bede-0e7b3da7bab6
    """

    created_at = table_column.DateTimeField(auto_now_add=True)
    """
    timestamping for when this was added to the database
    """

    updated_at = table_column.DateTimeField(auto_now=True)
    """
    timestamping for when this was last updated
    """

    corrections = table_column.JSONField(blank=True, null=True)
    """
    Simmate workflows often have ErrorHandlers that fix any issues while the
    calaculation ran. This often involves changing settings, so we store
    any of those changes here.
    """

    """Archived Data"""
    # I may want to add these fields because Prefect doesn't store run stats
    # indefinitely AND it doesn't give detail memory use, number of cores, etc.
    # If all of these are too much, I could do a json field of "run_stats" instead

    # average_memory (The average memory used in kb)
    # max_memory (The maximum memory used in kb)
    # elapsed_time (The real time elapsed in seconds)
    # system_time(The system CPU time in seconds)
    # user_time(The user CPU time spent by VASP in seconds)
    # total_time (The total CPU time for this calculation)
    # cores (The number of cores used by VASP (some clusters print `mpi-ranks` here))

    # Here are some other Prefect fields too.
    # agent_id
    # auto_scheduled
    # context
    # created_by_user_id
    # end_time
    # flow_id
    # labels
    # name
    # parameters
    # scheduled_start_time
    # start_time
    # state
    # tenant_id
    # times_resurrected
    # version

    """ Relationships """
    # While there are no base relations for this abstract class, the majority of
    # calculations will link to a structure or alternatively a diffusion pathway
    # or crystal surface. In such cases, you'll add a relationship like this:
    #
    #   structure = table_column.ForeignKey(
    #       ExampleStructureModel,
    #       on_delete=table_column.CASCADE,
    #   )
    #
    # Note that this is a ForeignKey because we want to allow the user to attempt
    # the same calculation multiple times. Otherwise it would be a OneToOneField.

    """ Properties """

    @property
    def prefect_cloud_link(self) -> str:
        """
        URL to this calculation (flow-run) in the Prefect Cloud webstite.
        """
        return f"https://cloud.prefect.io/simmate/flow-run/{self.prefect_flow_run_id}"

    """ Model Methods """

    @classmethod
    def from_prefect_id(cls, id: str, **kwargs):
        """
        Given a prefect id, this method will do one of the following...

        1. if it exists, load the database object with matching prefect_id
        2. if it doesn't exist, create a new database object using this ID and extra kwargs

        """

        # Depending on how a workflow was submitted, there may be a calculation
        # extry existing already -- which we need to grab and then update. If it's
        # not there, we create a new one!

        # check if the calculation already exists in our database, and if so,
        # grab it and return it.
        if cls.objects.filter(prefect_flow_run_id=id).exists():
            return cls.objects.get(prefect_flow_run_id=id)
        # Otherwise we need to create a new one and return that.

        # See if the calculation table has a from_toolkit method (most simmate calcs do)
        # and if so, we should call that instead and pass all of our kwargs to it.
        if hasattr(cls, "from_toolkit"):
            calculation = cls.from_toolkit(prefect_flow_run_id=id, **kwargs)
        else:
            calculation = cls(prefect_flow_run_id=id, **kwargs)
        calculation.save()

        return calculation



# -*- coding: utf-8 -*-

"""
This module is experimental and subject to change.
"""


from pymatgen.core.sites import PeriodicSite
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from simmate.toolkit.diffusion import MigrationHop as ToolkitMigrationHop
from simmate.database.base_data_types import (
    table_column,
    DatabaseTable,
    Structure,
)


class DiffusionAnalysis(Structure):
    class Meta:
        abstract = True
        app_label = "local_calculations"

    # The element of the diffusion atom
    migrating_specie = table_column.CharField(max_length=4, blank=True, null=True)

    # Whether vacancy or interstitial diffusion was used
    vacancy_mode = table_column.BooleanField(blank=True, null=True)

    # atomic fraction of the diffusion ion
    atomic_fraction = table_column.FloatField(blank=True, null=True)

    # Evaluates all MigrationHops to find the lowest-barrier percolation network
    barrier_cell = table_column.FloatField(blank=True, null=True)
    paths_involved = table_column.CharField(max_length=100, blank=True, null=True)
    npaths_involved = table_column.IntegerField(blank=True, null=True)

    @classmethod
    def from_toolkit(
        cls,
        migrating_specie: str = None,
        vacancy_mode: bool = None,
        as_dict: bool = False,
        **kwargs,
    ):
        # the algorithm doesn't change for this method, but we do want to add
        # a few extra columns. Therefore we make the dictionary as normal and
        # then add those extra columns here.
        structure_dict = super().from_toolkit(as_dict=True, **kwargs)

        # now add the two extra columns
        structure_dict["migrating_specie"] = migrating_specie
        # Structure should be present -- otherwise an error would have been
        # raised by the call to super()
        structure_dict["atomic_fraction"] = (
            kwargs["structure"].composition.get_atomic_fraction(migrating_specie)
            if migrating_specie
            else None
        )

        # If as_dict is false, we build this into an Object. Otherwise, just
        # return the dictionary
        return structure_dict if as_dict else cls(**structure_dict)

    @classmethod
    def create_subclasses(
        cls,
        name: str,
        module: str,
        **extra_columns,
    ):
        """
        Dynamically creates a subclass of Relaxation as well as a separate IonicStep
        table for it. These tables are linked together.

        Example use:

        ``` python
        from simmate.database.base_data_types import Relaxation

        # note the odd formatting here is just because we are parsing three
        # outputs from this method to three variables.
        (
            ExampleDiffusionAnalysis,
            ExampleMigrationHop,
            ExampleMigrationImage,
        ) = DiffusionAnalysis.create_subclasses(
            "Example",
            module=__name__,
        )
        ```

        Parameters
        ----------
        - `name` :
            The prefix name of the subclasses that are output. "Relaxation" and
            "IonicStep" will be attached to the end of this prefix.

        - `module` :
            name of the module this subclass should be associated with. Typically,
            you should pass __name__ to this.

        - `**extra_columns` :
            Additional columns to add to the table. The keyword will be the
            column name and the value should match django options
            (e.g. table_column.FloatField())

        Returns
        -------
        - `NewDiffusionAnalysisClass` :
            A subclass of DiffusionAnalysis.
        - `NewMigrationHopClass`:
            A subclass of MigrationHop.
        - `NewMigrationHopClass`:
            A subclass of MigrationHop.
        """

        # For convience, we add columns that point to the start and end structures
        NewDiffusionAnalysisClass = cls.create_subclass(
            f"{name}DiffusionAnalysis",
            module=module,
            **extra_columns,
        )

        (
            NewMigrationHopClass,
            NewMigrationImageClass,
        ) = MigrationHop.create_subclasses_from_diffusion_analysis(
            name,
            NewDiffusionAnalysisClass,
            module=module,
            **extra_columns,
        )

        # we now have a new child class and avoided writing some boilerplate code!
        return NewDiffusionAnalysisClass, NewMigrationHopClass, NewMigrationImageClass


# TODO: consider making a Calculation bc this is what the corrections/directory
# information should be attached to.
class MigrationHop(DatabaseTable):
    class Meta:
        abstract = True
        app_label = "local_calculations"

    # the initial, midpoint, and end site fractional coordinates
    # Really, this is a list of float values, but I save it as a string.
    # !!! for robustness, should I save cartesian coordinates and/or lattice as well?
    # !!! Does the max length make sense here and below?
    # !!! I could also store as JSON since it's a list of coords.
    site_start = table_column.CharField(max_length=100, blank=True, null=True)
    site_end = table_column.CharField(max_length=100, blank=True, null=True)

    # BUG: the init script for MigrationHop can't identify the site indexes
    # properly but they should be the same as before because it is a symmetrized
    # structure. Note that even if I'm wrong in some case -- this will have
    # no effect because iindex and eindex are only used in one portion of
    # the hash as well as for printing the __str__ of the object.
    index_start = table_column.IntegerField(blank=True, null=True)
    index_end = table_column.IntegerField(blank=True, null=True)

    """ Query-helper Info """

    # TODO:
    # The expected index in DistinctPathFinder.get_paths. The shortest path is index 0
    # and they are all ordered by increasing length.
    number = table_column.IntegerField(blank=True, null=True)

    # The length/distance of the pathway from start to end (linear measurement)
    length = table_column.FloatField(blank=True, null=True)

    # pathway dimensionality
    dimension_path = table_column.IntegerField(blank=True, null=True)
    dimension_host_lattice = table_column.IntegerField(blank=True, null=True)

    # Evaluates all MigrationImages to find the barriers
    energy_barrier = table_column.FloatField(blank=True, null=True)

    # TODO:
    # Distance of the pathway relative to the shortest pathway distance
    # in the structure using the formula: (D - Dmin)/Dmin
    # distance_rel_min = models.FloatField()

    """ Relationships """
    # Each MigrationHop corresponds to the full analysis of one Structure, which
    # can have many MigrationHop(s)
    # diffusion_analysis = table_column.ForeignKey(
    #     # DiffusionAnalysis,
    #     on_delete=table_column.PROTECT,
    #     related_name="migration_hops",
    # )

    # TODO:
    # image_start --> OneToOneField for specific MigrationHop
    # image_end --> OneToOneField for specific MigrationHop
    # image_transition_state --> OneToOneField for specific MigrationHop

    # Just like Relaxation points to IonicSteps, NEB will point to MigrationImages

    """ Model Methods """
    # TODO: If I want a queryset to return a pymatgen-diffusion object(s) directly,
    # then I need make a new Manager rather than adding methods here.

    @classmethod
    def _from_toolkit(
        cls,
        migration_hop: ToolkitMigrationHop,
        as_dict: bool = False,
        number: int = None,
        **kwargs,
    ):

        # convert the pathway object into the database table format
        hop_dict = dict(
            site_start=" ".join(str(c) for c in migration_hop.isite.frac_coords),
            site_end=" ".join(str(c) for c in migration_hop.esite.frac_coords),
            index_start=migration_hop.iindex,
            index_end=migration_hop.eindex,
            length=migration_hop.length,
            # diffusion_analysis_id=diffusion_analysis_id,
            **kwargs,
        )

        # If as_dict is false, we build this into an Object. Otherwise, just
        # return the dictionary
        return hop_dict if as_dict else cls(**hop_dict)

    # BUG: because of rounding in the from_toolkit method, the get_sc_structures
    # is unable to identify equivalent sites. I opened an issue for this
    # with their team:
    #   https://github.com/materialsvirtuallab/pymatgen-analysis-diffusion/issues/296
    def to_toolkit(self):
        """
        converts the database MigrationHop to a toolkit MigrationHop
        """
        # The bulk crystal structure is stored in the diffusion analysis table
        structure = self.diffusion_analysis.to_toolkit()

        # pathways require a symmetrized structure
        # BUG: I need to ensure the symmetrized structure comes out the same
        # as before, so I need to be careful with updates and hardcoding here.
        sga = SpacegroupAnalyzer(structure, symprec=0.1)
        symm_structure = sga.get_symmetrized_structure()

        isite_new = PeriodicSite(
            species=self.diffusion_analysis.migrating_specie,
            coords=[float(x) for x in self.site_start.split(" ")],
            lattice=symm_structure.lattice,
        )
        esite_new = PeriodicSite(
            species=self.diffusion_analysis.migrating_specie,
            coords=[float(x) for x in self.site_end.split(" ")],
            lattice=symm_structure.lattice,
        )

        path = ToolkitMigrationHop(isite_new, esite_new, symm_structure)
        # BUG: see comment attached to definition of these fields
        path.iindex = self.index_start
        path.eindex = self.index_end

        # if the pathways match, then we can return the pathway object!
        return path

    @classmethod
    def create_subclasses_from_diffusion_analysis(
        cls,
        name: str,
        diffusion_analysis: DiffusionAnalysis,
        module: str,
        **extra_columns,
    ):
        """
        Dynamically creates subclass of MigrationHop and MigrationImage, then
        links them to the DiffusionAnalysis table.

        This method should NOT be called directly because it is instead used by
        `DiffusionAnalysis.create_subclasses`.

        Parameters
        ----------
        - `name` :
            Name of the subclass that is output.
        - `diffusion_analysis` :
            DiffusionAnalysis table that these subclasses should be associated with.
        - `module` :
            name of the module this subclass should be associated with. Typically,
            you should pass __name__ to this.
        - `**extra_columns` :
            Additional columns to add to the table. The keyword will be the
            column name and the value should match django options
            (e.g. table_column.FloatField())

        Returns
        -------
        - `NewMigrationHopClass`:
            A subclass of MigrationHop.
        - `NewMigrationHopClass`:
            A subclass of MigrationHop.
        """

        NewMigrationHopClass = cls.create_subclass(
            f"{name}MigrationHop",
            diffusion_analysis=table_column.ForeignKey(
                diffusion_analysis,
                on_delete=table_column.CASCADE,
                related_name="migration_hops",
            ),
            module=module,
            **extra_columns,
        )

        NewMigrationImageClass = MigrationImage.create_subclass_from_migration_hop(
            name,
            NewMigrationHopClass,
            module=module,
            **extra_columns,
        )

        return NewMigrationHopClass, NewMigrationImageClass


class MigrationImage(Structure):
    class Meta:
        abstract = True
        app_label = "local_calculations"

    base_info = [
        "number",
        "structure_string",
        "force_tangent",
        "structure_distance",
        "energy",
    ]

    # 0 = start, -1 = end.
    number = table_column.IntegerField()

    # Diffusion analysis given a tangent force, so we don't mix this up with the
    # Force mix-in.
    force_tangent = table_column.FloatField(blank=True, null=True)

    # For NEB, we only care about the total energy -- not the other fields that
    # the Thermodynamics mix-in provides.
    energy = table_column.FloatField(blank=True, null=True)

    # This measures the fingerprint distance of the image from the starting image
    structure_distance = table_column.FloatField(blank=True, null=True)

    # We don't need the source column for the MigrationImage class because we
    # instead stored on the DiffusionAnalysis object. This line deletes the
    # source could from our Structure mix-in.
    source = None

    def update_many_from_analysis(self):
        # StaticEnergy.update_from_vasp_run but adjusted to load a NEB vasprun
        pass

    @classmethod
    def create_subclass_from_migration_hop(
        cls,
        name: str,
        migration_hop: MigrationHop,
        module: str,
        **extra_columns,
    ):
        """
        Dynamically creates subclass of MigrationImage, then links it to the
        MigrationHop table.

        This method should NOT be called directly because it is instead used by
        `DiffusionAnalysis.create_subclasses`.

        Parameters
        ----------
        - `name` :
            Name of the subclass that is output.
        - `migration_hop` :
            MigrationHop table that these images should be associated with.
        - `module` :
            name of the module this subclass should be associated with. Typically,
            you should pass __name__ to this.
        - `**extra_columns` :
            Additional columns to add to the table. The keyword will be the
            column name and the value should match django options
            (e.g. table_column.FloatField())

        Returns
        -------
        - `NewClass` :
            A subclass of MigrationImage.

        """

        NewClass = cls.create_subclass(
            f"{name}MigrationImage",
            migration_hop=table_column.ForeignKey(
                migration_hop,
                on_delete=table_column.CASCADE,
                related_name="migration_images",
            ),
            module=module,
            **extra_columns,
        )

        return NewClass



# -*- coding: utf-8 -*-

import click

from typing import List


@click.group()
def workflows():
    """A group of commands for running workflows or viewing their settings."""
    pass


def get_workflow(workflow_name: str):
    """
    This is a utility for click (cli) that grabs a workflow from the simmate
    workflows. If the workflow can't be found, it raises a ClickException.

    Parameters
    ----------
    workflow_name : str
        Name of the workflow to grab (e.g. relaxation-matproj)
    """

    from simmate.workflows.utilities import get_list_of_all_workflows

    allowed_workflows = get_list_of_all_workflows()

    # make sure we have a proper workflow name provided
    if workflow_name not in allowed_workflows:
        raise click.ClickException(
            "The workflow you provided isn't known. Make sure you don't have any "
            "typos! If you want a list of all available workflows, use the command "
            "`simmate workflows list-all`. You can also interactively explore "
            "workflows with `simmate workflows explore`"
        )

    from importlib import import_module

    # parse the workflow name. (e.g. static-energy/mit --> static-energy + mit)
    type_name, preset_name = workflow_name.split("/")
    type_name = type_name.replace("-", "_")
    preset_name = preset_name.replace("-", "_")

    # The naming convention matches the import path, so we can load the workflow
    workflow_module = import_module(f"simmate.workflows.{type_name}")
    print(f"simmate.workflows.{type_name}")
    print(f"{preset_name}_workflow")
    workflow = getattr(workflow_module, f"{preset_name}_workflow")

    return workflow


def list_options(options: List) -> int:
    """
    This is a utility for click (cli) that prints of list of items as a numbered
    list. It prompts users to select an option from the list.

    For example, `["item1", "item2", "item3"]` would print...
    ```
        (01) item1
        (02) item2
        (03) item3
    ```

    Parameters
    ----------
    - `options`:
        a list of strings to choose from

    Returns
    --------
    - `selected_index`:
        The integer value of the choice selected. This will follw python indexing
        so the index of the options list. (e.g. if item1 was selected, 0 would be
        returned)

    """

    for i, item in enumerate(options):
        number = str(i + 1).zfill(2)
        click.echo(f"\t({number}) {item}")

    # Have the user select an option. We use -1 because indexing count is from 0, not 1.
    selected_index = click.prompt("\n\nPlease choose a number:", type=int) - 1

    if selected_index >= len(options) or selected_index < 0:
        raise click.ClickException(
            "Number does not match any the options provided. Exiting."
        )

    click.echo(f"You have selectd `{options[selected_index]}`.")

    return selected_index


@workflows.command()
def explore():
    """
    Let's you interactively view all available workflows and see the documentation
    on the one you select.
    """

    click.echo("\nGathering all available workflows...")
    from simmate.workflows.utilities import (
        ALL_WORKFLOW_TYPES,
        get_list_of_workflows_by_type,
    )

    click.echo("\n\nWhat type of analysis are you interested in?")
    types_cleaned = [t.replace("_", " ") for t in ALL_WORKFLOW_TYPES]
    type_index = list_options(types_cleaned)
    selected_type = ALL_WORKFLOW_TYPES[type_index]

    # TODO: have the user select a calculator for this analysis. For now,
    # we are assuming VASP because those are our only workflows

    click.echo("\n\nWhat settings preset do you want to see the description for?")
    presets = [t for t in get_list_of_workflows_by_type(selected_type)]
    present_index = list_options(presets)
    selected_preset = presets[present_index]

    final_workflow_name = selected_type.replace("_", "-") + "/" + selected_preset
    click.echo(f"\n\n===================== {final_workflow_name} =====================")

    # now we load this workflow and print the docstring.
    workflow = get_workflow(final_workflow_name)

    click.echo(workflow.__doc__)

    click.echo("==================================================================")


@workflows.command()
def list_all():
    """
    This lists off all available workflows.
    """

    click.echo("Gathering all available workflows...")

    from simmate.workflows.utilities import get_list_of_all_workflows

    click.echo("These are the workflows that have been registerd:")
    all_workflows = get_list_of_all_workflows()
    for i, workflow in enumerate(all_workflows):
        # Replace underscores with dashes for consistency with click
        workflow_dash = workflow.replace("_", "-")
        workflow_number = str(i + 1).zfill(2)
        click.echo(f"\t({workflow_number}) {workflow_dash}")  # gives "(1) example-flow"


@workflows.command()
@click.argument("workflow_name")
def show_config(workflow_name):
    """
    If the workflow is a single task, the calculation's configuration settings
    are displayed. For example, a VASP workflow will show a dictionary that
    details how INCAR settings are selected.
    """

    click.echo("LOADING WORKFLOW...")
    workflow = get_workflow(workflow_name)

    click.echo("PRINTING WORKFLOW CONFIG...")

    # Not all workflows have a single config because some are NestWorkflows,
    # meaning they are made of multiple smaller workflows.
    if hasattr(workflow, "s3task"):
        workflow.s3task.print_config()
    elif hasattr(workflow, "s3tasks"):
        raise click.ClickException(
            "This is a NestedWorkflow, meaning it is made up of multiple smaller "
            "workflows. We have not added a show-config feature for these yet. "
            "This will be added before version 0.0.0 though."
        )


@workflows.command()
@click.argument("workflow_name")
@click.argument("filename", type=click.Path(exists=True))
@click.option(
    "--directory",
    "-d",
    default=None,
    help="the folder to write input file in. Defaults to <workflow_name>_input",
)
def setup_only(workflow_name, filename, directory):
    """
    If the workflow is a single task, the calculation is set up but not ran. This
    is useful when you just want the input files to view/edit.
    """

    click.echo("LOADING STRUCTURE AND WORKFLOW...")
    from simmate.toolkit import Structure

    workflow = get_workflow(workflow_name)
    structure = Structure.from_file(filename)

    click.echo("WRITING INPUT FILES...")

    # if no folder was given, just name it after the workflow. We also replace
    # spaces with underscores
    from simmate.utilities import get_directory

    if not directory:
        directory = get_directory(f"{workflow.name}_inputs".replace(" ", "_"))

    # Not all workflows have a single input because some are NestWorkflows,
    # meaning they are made of multiple smaller workflows.
    if hasattr(workflow, "s3task"):
        workflow.s3task().setup(structure, directory)
    elif hasattr(workflow, "s3tasks"):
        raise click.ClickException(
            "This is a NestedWorkflow, meaning it is made up of multiple smaller "
            "workflows. We have not added a setup-only feature for these yet."
        )

    click.echo(f"Done! Your input files are located in {directory}")


# These context settings let us pass extra kwargs, as some workflows have unique
# inputs (e.g. structures, migration_hop, etc. instead of a structure input).
# This is also why we have pass_context and then the kwarg context.
@workflows.command(
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    )
)
@click.argument("workflow_name")
@click.option(
    "--structure",
    "-s",
    default=None,
    type=click.Path(exists=True),
    help="filename of the structure used for this workflow (note, not all workflows use this arg)",
)
@click.option(
    "--command",
    "-c",
    default=None,
    help="the command used to call call the calculator (e.g. 'mpirun -n 12 vasp_std > vasp.out')",
)
@click.option(
    "--directory",
    "-d",
    default=None,
    help="the folder to run this workflow in. Defaults to simmate-task-12345, where 12345 is randomized",
)
@click.pass_context
def run(context, workflow_name, structure, command, directory):
    """Runs a workflow using an input structure file"""

    click.echo("LOADING WORKFLOW & INPUT PARAMETERS...")
    from simmate.toolkit import Structure

    workflow = get_workflow(workflow_name)

    # ---- Parsing parameters ----

    # we don't want to pass arguments like command=None or structure=None if the
    # user didn't provide this input. Instead, we want the workflow to use
    # its own default value. To do this, we pass the command input as a kwarg
    # or an empty dict if it wasn't given.

    # The extra keywords passed are given as a list such as...
    #   ["--arg1", "val1", "--arg2", "val2",]
    # Using code from stack overflow, we parse that data into a dictionary.
    #   https://stackoverflow.com/questions/32944131/
    # If there's any issue parsing the data, we alert the user with a hint.
    try:
        extra_kwargs = {
            context.args[i][2:]: context.args[i + 1]
            for i in range(0, len(context.args), 2)
        }
    except:
        raise click.ClickException(
            "This workflow command appears to be improperly formatted. \n"
            "When giving workflow parameters, make sure you provide both the \n"
            "keyword and the value (such as `--example_kwarg example_value`)."
        )

    if command:
        extra_kwargs["command"] = command

    # The structure input (which is optional) is given as a filename, which
    # we need to load as toolkit object. This is also the case for a number
    # of inputs like migration_hop, structures, etc.
    if structure:
        extra_kwargs["structure"] = Structure.from_file(structure)

    if "structures" in extra_kwargs.keys():
        structure_filenames = extra_kwargs["structures"].split(";")
        extra_kwargs["structures"] = [
            Structure.from_file(file) for file in structure_filenames
        ]

    if "migration_hop" in extra_kwargs.keys():
        from simmate.toolkit.diffusion import MigrationHop

        migration_hop = MigrationHop.from_dynamic(extra_kwargs["migration_hop"])
        extra_kwargs["migration_hop"] = migration_hop

    if "supercell_start" in extra_kwargs.keys():
        extra_kwargs["supercell_start"] = Structure.from_file(
            extra_kwargs["supercell_start"]
        )

    if "supercell_end" in extra_kwargs.keys():
        extra_kwargs["supercell_end"] = Structure.from_file(
            extra_kwargs["supercell_end"]
        )

    # ---- end of parsing parameters

    click.echo("RUNNING WORKFLOW...")

    result = workflow.run(
        directory=directory,
        **extra_kwargs,
    )

    # Let the user know everything succeeded
    # if result.is_successful():
    #     click.echo("Success! All results are also stored in your database.")
    # BUG: we remove this temporarily because we have s3tasks ran through this
    # module. When we add their database components, this will be removed.


@workflows.command()
@click.argument("workflow_name")
@click.argument("filename", type=click.Path(exists=True))
@click.option(
    "--command",
    "-c",
    default=None,
    help="the command used to call call the calculator (e.g. 'mpirun -n 12 vasp_std > vasp.out')",
)
@click.option(
    "--directory",
    "-d",
    default=None,
    help="the folder to run this workflow in. Defaults to simmate-task-12345, where 12345 is randomized",
)
def run_cloud(workflow_name, filename, command, directory):
    """Runs a workflow using an input structure file"""

    click.echo("LOADING STRUCTURE AND WORKFLOW...")
    from simmate.toolkit import Structure

    workflow = get_workflow(workflow_name)
    structure = Structure.from_file(filename)

    click.echo("RUNNING WORKFLOW...")

    # we don't want to pass command=None if the user didn't provide one. Instead
    # we want the workflow to use its own default value. To do this, we pass
    # the command input as a kwarg -- or an empty dict if it wasn't given.
    command_kwargs = {"command": command} if command else {}

    result = workflow.run_cloud(
        structure=structure,
        directory=directory,
        **command_kwargs,
    )

    # Let the user know everything succeeded
    if result.is_successful():
        click.echo("Success! All results are also stored in your database.")
